# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: _13_create_hyrex_task_run_table_indexes.sql
import dataclasses

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


CREATE_TASK_RUN_TABLE_INDEXES = """-- name: create_task_run_table_indexes \\:exec
DO $$
BEGIN
    -- Create indexes for hyrex_task_run table
    CREATE INDEX IF NOT EXISTS ix_hyrex_task_run_task_name
        ON public.hyrex_task_run (task_name);

    CREATE INDEX IF NOT EXISTS ix_hyrex_task_run_status
        ON public.hyrex_task_run (status);

    CREATE INDEX IF NOT EXISTS ix_hyrex_task_run_queue
        ON public.hyrex_task_run (queue);

    CREATE INDEX IF NOT EXISTS ix_hyrex_task_run_scheduled_start
        ON public.hyrex_task_run (scheduled_start);

    CREATE INDEX IF NOT EXISTS index_queue_status
        ON public.hyrex_task_run (status, queue, scheduled_start, task_name);

    CREATE UNIQUE INDEX IF NOT EXISTS ix_hyrex_task_run_idempotency_key
        ON public.hyrex_task_run (task_name, idempotency_key)
        WHERE idempotency_key IS NOT NULL;

    CREATE INDEX IF NOT EXISTS idx_hyrex_task_run_queue_status_priority_queued
        ON hyrex_task_run (queue, status, priority ASC, queued);

    CREATE INDEX IF NOT EXISTS idx_htr_queued_started_finished
        ON hyrex_task_run(queued, started, finished);

    CREATE INDEX IF NOT EXISTS idx_hyrex_task_run_root_id
        ON hyrex_task_run(root_id);
END $$
"""


@dataclasses.dataclass()
class CreateTaskRunTableIndexesParams:

    pass

class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def create_task_run_table_indexes(self, arg: CreateTaskRunTableIndexesParams) -> None:
        self._conn.execute(sqlalchemy.text(CREATE_TASK_RUN_TABLE_INDEXES))


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def create_task_run_table_indexes(self, arg: CreateTaskRunTableIndexesParams) -> None:
        await self._conn.execute(sqlalchemy.text(CREATE_TASK_RUN_TABLE_INDEXES))



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: register_app_info.sql
import dataclasses
from typing import Any, Optional

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


REGISTER_APP_INFO = """-- name: register_app_info \\:exec
INSERT INTO hyrex_app (
    id,
    app_info
) VALUES (
    :p1,
    :p2
)
ON CONFLICT (id) DO UPDATE SET
    app_info = :p2
"""


@dataclasses.dataclass()
class RegisterAppInfoParams:
    id: int
    app_info: Optional[Any]


class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def register_app_info(self, arg: RegisterAppInfoParams) -> None:
        self._conn.execute(sqlalchemy.text(REGISTER_APP_INFO), {"p1": arg.id, "p2": arg.app_info})


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def register_app_info(self, arg: RegisterAppInfoParams) -> None:
        await self._conn.execute(sqlalchemy.text(REGISTER_APP_INFO), {"p1": arg.id, "p2": arg.app_info})



import logging

from .hyrex_app import HyrexApp
from .hyrex_cache import HyrexCache
from .hyrex_context import HyrexContext, get_hyrex_context, get_hyrex_workflow_context
from .hyrex_kv import HyrexKV
from .hyrex_queue import HyrexQueue
from .hyrex_registry import HyrexRegistry

# Set up default logging configuration
logger = logging.getLogger("hyrex")
logger.setLevel(logging.INFO)
# Add a console handler if none exists
if not logger.handlers:
    handler = logging.StreamHandler()
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    handler.setFormatter(formatter)
    logger.addHandler(handler)



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: _15_create_hyrex_kv_table.sql
import dataclasses

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


CREATE_HYREX_KV_TABLE = """-- name: create_hyrex_kv_table \\:exec
CREATE TABLE IF NOT EXISTS hyrex_kv (
    key VARCHAR NOT NULL PRIMARY KEY,
    value TEXT NOT NULL,
    created_at TIMESTAMP WITH TIME ZONE NOT NULL DEFAULT CURRENT_TIMESTAMP
)
"""


@dataclasses.dataclass()
class CreateHyrexKvTableParams:

    pass

class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def create_hyrex_kv_table(self, arg: CreateHyrexKvTableParams) -> None:
        self._conn.execute(sqlalchemy.text(CREATE_HYREX_KV_TABLE))


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def create_hyrex_kv_table(self, arg: CreateHyrexKvTableParams) -> None:
        await self._conn.execute(sqlalchemy.text(CREATE_HYREX_KV_TABLE))



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: _04_create_hyrex_scheduler_lock_table.sql
import dataclasses

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


CREATE_SCHEDULER_LOCK_TABLE = """-- name: create_scheduler_lock_table \\:exec
CREATE TABLE IF NOT EXISTS hyrex_scheduler_lock
(
    lockid       bigserial PRIMARY KEY,
    worker_name  text        NOT NULL,
    acquired_at  timestamptz NOT NULL DEFAULT now(),
    heartbeat_at timestamptz NOT NULL DEFAULT now(),
    release_at   timestamptz NOT NULL,
    is_active    boolean     NOT NULL DEFAULT true
)
"""


@dataclasses.dataclass()
class CreateSchedulerLockTableParams:

    pass

class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def create_scheduler_lock_table(self, arg: CreateSchedulerLockTableParams) -> None:
        self._conn.execute(sqlalchemy.text(CREATE_SCHEDULER_LOCK_TABLE))


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def create_scheduler_lock_table(self, arg: CreateSchedulerLockTableParams) -> None:
        await self._conn.execute(sqlalchemy.text(CREATE_SCHEDULER_LOCK_TABLE))



from typing import Type
from uuid import UUID

from pydantic import BaseModel
from uuid6 import uuid7

from hyrex.configs import ConfigPhase, TaskConfig, WorkflowConfig
from hyrex.dispatcher.dispatcher import Dispatcher
from hyrex.hyrex_queue import HyrexQueue
from hyrex.schemas import EnqueueTaskRequest, TaskStatus, WorkflowRunRequest
from hyrex.workflow.workflow_builder import DagNode, WorkflowBuilder


class HyrexWorkflow:
    def __init__(
        self,
        name: str,
        workflow_config: WorkflowConfig,
        workflow_arg_schema: Type[BaseModel] | None,
        workflow_builder: WorkflowBuilder,
        dispatcher: Dispatcher,
        source_code: str,
        cron: str | None,
    ):
        self.name = name
        self.workflow_config = workflow_config
        self.workflow_arg_schema = workflow_arg_schema
        self.workflow_builder = workflow_builder
        self.dispatcher = dispatcher
        self.source_code = source_code
        self.cron = cron

    def get_queue(self) -> HyrexQueue | str | None:
        return self.workflow_config.queue

    def with_config(
        self, queue: str | HyrexQueue = None, priority: int = None
    ) -> "HyrexWorkflow":
        new_workflow_config = WorkflowConfig(
            config_phase=ConfigPhase.send, queue=queue, priority=priority
        )
        new_workflow = HyrexWorkflow(
            name=self.name,
            workflow_config=self.workflow_config.merge(new_workflow_config),
            workflow_arg_schema=self.workflow_arg_schema,
            workflow_builder=self.workflow_builder,
            dispatcher=self.dispatcher,
        )
        return new_workflow

    def serialize_workflow_to_task_requests(
        self,
        workflow_run_id: UUID,
    ) -> list[EnqueueTaskRequest]:
        node_to_task_request: dict[DagNode, EnqueueTaskRequest] = {}
        dependency_map: dict[DagNode, set[UUID]] = {}

        for node in self.workflow_builder.get_root_nodes():
            self.traverse(
                node=node,
                node_to_request=node_to_task_request,
                dependency_map=dependency_map,
                workflow_run_id=workflow_run_id,
            )

        for node, task_request in node_to_task_request.items():
            deps = dependency_map.get(node)
            if deps and len(deps) > 0:
                task_request.workflow_dependencies = list(deps)
            else:
                # Root nodes (no dependencies) should be queued immediately
                task_request.status = TaskStatus.queued

        return node_to_task_request.values()

    def send(self, context: BaseModel):
        # Runtime type checking to ensure context is the expected schema type
        if not isinstance(context, self.workflow_arg_schema):
            raise TypeError(
                f"Expected context of type {self.workflow_arg_schema.__name__}, "
                f"got {type(context).__name__} instead"
            )

        workflow_run_request = WorkflowRunRequest(
            id=uuid7(),
            workflow_name=self.name,
            args=context.model_dump(),
            queue=self.workflow_config.queue,
            timeout_seconds=self.workflow_config.timeout_seconds,
            idempotency_key=self.workflow_config.idempotency_key,
        )

        self.dispatcher.send_workflow_run(workflow_run_request)

    def traverse(
        self,
        node: DagNode,
        node_to_request: dict[DagNode, EnqueueTaskRequest],
        dependency_map: dict[DagNode, set[UUID]],
        workflow_run_id: UUID,
    ) -> None:
        """
        Recursively traverses the workflow DAG starting from the specified node and serializes each node into a task request.

        For each node, a unique UUID is generated which is used as the task's `id`, `durable_id`, and `root_id`.
        The workflow run ID is assigned to each task request, and for every child node, the parent's ID is added as a dependency.

        Args:
            node: The current workflow node to process.
            node_to_request: A map tracking nodes that have already been serialized along with their corresponding task requests.
            dependency_map: A map recording dependency edges, mapping each node to a set of parent task IDs (dependencies).
            workflow_run_id: The workflow run ID to assign to every task request.
        """
        # Create a new task request if the node hasn't been processed
        if node not in node_to_request:
            # Generate a single UUID for the task
            task_id = uuid7()

            # For max_retries, timeout_seconds, idempotency_key, etc.
            task_config = node.task_wrapper.task_config.apply_workflow_config(
                self.workflow_config
            )

            # All three IDs are set to the same UUID
            task_request = EnqueueTaskRequest(
                id=task_id,
                durable_id=task_id,
                root_id=task_id,
                workflow_run_id=workflow_run_id,
                workflow_dependencies=None,
                parent_id=None,
                status=TaskStatus.await_deps,
                task_name=node.task_wrapper.task_identifier,
                args={},
                queue=task_config.get_queue_name(),
                max_retries=task_config.max_retries,
                priority=task_config.priority,
                timeout_seconds=task_config.timeout_seconds,
                idempotency_key=task_config.idempotency_key,
            )

            node_to_request[node] = task_request

        current_request = node_to_request[node]

        # Process each child of the current node
        for child in node.get_children():
            # Record the dependency: the current task's id (a.k.a. durable_id) is a prerequisite for the child
            if child not in dependency_map:
                dependency_map[child] = set()
            dependency_map[child].add(current_request.id)

            # Recursively traverse the child node
            self.traverse(child, node_to_request, dependency_map, workflow_run_id)



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: create_transition_task_run_state_func.sql
import dataclasses

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


CREATE_TRANSITION_TASK_RUN_STATE_FUNC = """-- name: create_transition_task_run_state_func \\:exec
DO $$
BEGIN
    -- ──────────────────────────────────────────────────────────────
    --  is_allowed(from_state, to_state)
    --  • Returns TRUE iff the edge exists in the Hyrex task-run FSM
    --  • Immutable, so PostgreSQL inlines it into plans (O-(1) test)
    -- ──────────────────────────────────────────────────────────────
    CREATE OR REPLACE FUNCTION is_allowed(
        p_from task_run_status,
        p_to   task_run_status
    ) RETURNS boolean
        IMMUTABLE
        LANGUAGE sql AS $func$
    SELECT (p_from, p_to) IN (
        -- ──────────────────────
        -- AwaitDeps transitions
        -- ──────────────────────
        ROW('AWAIT_DEPS',       'AWAIT_START_TIME'),
        ROW('AWAIT_DEPS',       'CANCELED'),
        ROW('AWAIT_DEPS',       'QUEUED'),
        ROW('AWAIT_DEPS',       'SKIPPED'),

        -- ─────────────────────────────────
        -- AwaitStartTime transitions
        -- ─────────────────────────────────
        ROW('AWAIT_START_TIME', 'QUEUED'),
        ROW('AWAIT_START_TIME', 'CANCELED'),

        -- ────────────────
        -- Queued → …
        -- ────────────────
        ROW('QUEUED',           'RUNNING'),
        ROW('QUEUED',           'CANCELED'),

        -- ──────────────────────────────
        -- Running → terminal / cancel
        -- ──────────────────────────────
        ROW('RUNNING',          'STOPPED'),
        ROW('RUNNING',          'SUCCESS'),
        ROW('RUNNING',          'FAILED'),
        ROW('RUNNING',          'LOST'),
        ROW('RUNNING',          'UP_FOR_CANCEL'),

        -- ─────────────────────────
        -- UpForCancel → terminal
        -- ─────────────────────────
        ROW('UP_FOR_CANCEL',    'FAILED'),
        ROW('UP_FOR_CANCEL',    'CANCELED')
    );
    $func$;

    CREATE OR REPLACE FUNCTION transition_task_state(
        p_id        uuid,
        p_next      task_run_status           -- enum type
    ) RETURNS hyrex_task_run
        LANGUAGE plpgsql AS $func$
    DECLARE
        v_old  hyrex_task_run;        -- row before change
        v_new  hyrex_task_run;        -- row after  change
        v_error_msg text;
        v_error_detail text;
        v_error_hint text;
    BEGIN
        ------------------------------------------------------------------
        -- 1. Lock the row so concurrent transitions queue up.
        ------------------------------------------------------------------
        SELECT * INTO v_old
        FROM hyrex_task_run
        WHERE id = p_id
            FOR UPDATE;

        IF NOT FOUND THEN
            RAISE EXCEPTION 'task % not found', p_id;
        END IF;

        ------------------------------------------------------------------
        -- 2. Validate the transition against the FSM.
        ------------------------------------------------------------------
        IF NOT is_allowed(v_old.status, p_next) THEN      -- helper shown below
            RAISE EXCEPTION
                'illegal state change\\: % → %', v_old.status, p_next;
        END IF;

        ------------------------------------------------------------------
        -- 3. Apply the change + timestamp side-effects in one shot.
        ------------------------------------------------------------------
        UPDATE hyrex_task_run
        SET status   = p_next,
            queued   = CASE WHEN p_next = 'QUEUED'   THEN clock_timestamp() ELSE queued   END,
            started  = CASE WHEN p_next = 'RUNNING'  THEN clock_timestamp() ELSE started  END,
            finished = CASE WHEN p_next IN ('SUCCESS','FAILED',
                                            'LOST','STOPPED','SKIPPED','CANCELED')
                                THEN clock_timestamp()
                            ELSE finished
                END,
            last_heartbeat = clock_timestamp()
        WHERE id = p_id
        RETURNING * INTO v_new;

        RETURN v_new;

    EXCEPTION
        WHEN OTHERS THEN
            -- Capture error details
            GET STACKED DIAGNOSTICS 
                v_error_msg = MESSAGE_TEXT,
                v_error_detail = PG_EXCEPTION_DETAIL,
                v_error_hint = PG_EXCEPTION_HINT;
            
            -- Log the error to system logs
            INSERT INTO hyrex_system_logs (id, timestamp, event_name, event_body)
            VALUES (
                gen_random_uuid(),
                clock_timestamp(),
                'task_state_transition_failed',
                jsonb_build_object(
                    'task_id', p_id,
                    'from_state', v_old.status,
                    'to_state', p_next,
                    'error_message', v_error_msg,
                    'error_detail', v_error_detail,
                    'error_hint', v_error_hint,
                    'sqlstate', SQLSTATE
                )
            );
            
            -- Re-raise the original exception
            RAISE;
    END;
    $func$;
END $$
"""


@dataclasses.dataclass()
class CreateTransitionTaskRunStateFuncParams:

    pass

class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def create_transition_task_run_state_func(self, arg: CreateTransitionTaskRunStateFuncParams) -> None:
        self._conn.execute(sqlalchemy.text(CREATE_TRANSITION_TASK_RUN_STATE_FUNC))


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def create_transition_task_run_state_func(self, arg: CreateTransitionTaskRunStateFuncParams) -> None:
        await self._conn.execute(sqlalchemy.text(CREATE_TRANSITION_TASK_RUN_STATE_FUNC))



from abc import ABC, abstractmethod
from typing import Any, Hashable, TypeVar

_cache_manager: dict[Hashable, "HyrexCache"] = {}


class HyrexCacheManager:
    @staticmethod
    def get(key: Hashable) -> "HyrexCache | None":
        return _cache_manager.get(key)

    @staticmethod
    def set(key: Hashable, value: "HyrexCache") -> None:
        _cache_manager[key] = value

    @staticmethod
    def cleanup() -> None:
        for val in _cache_manager.values():
            val.cleanup(val.cached_object)


# Define a TypeVar
T = TypeVar("T")


class HyrexCache(ABC):
    """Base class for cached resources in Hyrex"""

    def __init__(self):
        self.cached_object: T | None = None

    @classmethod
    def get(cls, hashable_key: Hashable) -> T:
        instance = HyrexCacheManager.get(hashable_key)
        if instance:
            return instance.cached_object

        # Create new instance and initialize
        instance = cls()
        instance.cached_object = cls.initialize(hashable_key)
        HyrexCacheManager.set(hashable_key, instance)
        return instance.cached_object

    @staticmethod
    @abstractmethod
    def initialize(key: Hashable):
        pass

    @staticmethod
    @abstractmethod
    def cleanup(cached_object: T) -> None:
        pass



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: register_hype_cron_job.sql
import dataclasses
from typing import Any, Optional

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


REGISTER_HYPE_CRON_JOB = """-- name: register_hype_cron_job \\:exec
INSERT INTO hype_cron_job (jobname, schedule, command_type, command_params, active, should_backfill)
VALUES (:p1, :p2, :p3, :p4, :p5, :p6)
ON CONFLICT (jobname)
DO UPDATE SET 
    schedule = EXCLUDED.schedule,
    command_type = EXCLUDED.command_type,
    command_params = EXCLUDED.command_params,
    active = EXCLUDED.active,
    should_backfill = EXCLUDED.should_backfill
"""


@dataclasses.dataclass()
class RegisterHypeCronJobParams:
    jobname: str
    schedule: Optional[str]
    command_type: Any
    command_params: Any
    active: bool
    should_backfill: Optional[bool]


class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def register_hype_cron_job(self, arg: RegisterHypeCronJobParams) -> None:
        self._conn.execute(sqlalchemy.text(REGISTER_HYPE_CRON_JOB), {
            "p1": arg.jobname,
            "p2": arg.schedule,
            "p3": arg.command_type,
            "p4": arg.command_params,
            "p5": arg.active,
            "p6": arg.should_backfill,
        })


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def register_hype_cron_job(self, arg: RegisterHypeCronJobParams) -> None:
        await self._conn.execute(sqlalchemy.text(REGISTER_HYPE_CRON_JOB), {
            "p1": arg.jobname,
            "p2": arg.schedule,
            "p3": arg.command_type,
            "p4": arg.command_params,
            "p5": arg.active,
            "p6": arg.should_backfill,
        })



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: create_cron_job_for_sql_query.sql
import dataclasses
from typing import Optional

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


CREATE_CRON_JOB_FOR_SQL_QUERY = """-- name: create_cron_job_for_sql_query \\:exec
INSERT INTO hyrex_cron_job (schedule, command, jobname, should_backfill, job_source)
VALUES (:p1, :p2, :p3, :p4, 'SYSTEM')
ON CONFLICT (jobname) 
DO UPDATE SET 
    schedule = EXCLUDED.schedule,
    command = EXCLUDED.command,
    should_backfill = EXCLUDED.should_backfill,
    job_source = 'SYSTEM',
    active = true
"""


@dataclasses.dataclass()
class CreateCronJobForSqlQueryParams:
    schedule: Optional[str]
    command: str
    jobname: str
    should_backfill: Optional[bool]


class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def create_cron_job_for_sql_query(self, arg: CreateCronJobForSqlQueryParams) -> None:
        self._conn.execute(sqlalchemy.text(CREATE_CRON_JOB_FOR_SQL_QUERY), {
            "p1": arg.schedule,
            "p2": arg.command,
            "p3": arg.jobname,
            "p4": arg.should_backfill,
        })


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def create_cron_job_for_sql_query(self, arg: CreateCronJobForSqlQueryParams) -> None:
        await self._conn.execute(sqlalchemy.text(CREATE_CRON_JOB_FOR_SQL_QUERY), {
            "p1": arg.schedule,
            "p2": arg.command,
            "p3": arg.jobname,
            "p4": arg.should_backfill,
        })



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: advance_workflow_run_func.sql
import dataclasses
from typing import Any, AsyncIterator, Iterator, Optional
import uuid

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


ADVANCE_WORKFLOW_RUN_FUNC = """-- name: advance_workflow_run_func \\:many
SELECT advance_workflow_run FROM advance_workflow_run(:p1\\:\\:UUID)
"""


@dataclasses.dataclass()
class AdvanceWorkflowRunFuncParams:
    workflow_run_id: uuid.UUID


class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def advance_workflow_run_func(self, arg: AdvanceWorkflowRunFuncParams) -> Iterator[Optional[Any]]:
        result = self._conn.execute(sqlalchemy.text(ADVANCE_WORKFLOW_RUN_FUNC), {"p1": arg.workflow_run_id})
        for row in result:
            yield row[0]


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def advance_workflow_run_func(self, arg: AdvanceWorkflowRunFuncParams) -> AsyncIterator[Optional[Any]]:
        result = await self._conn.stream(sqlalchemy.text(ADVANCE_WORKFLOW_RUN_FUNC), {"p1": arg.workflow_run_id})
        async for row in result:
            yield row[0]



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: get_executors_paginated.sql
import dataclasses
import datetime
from typing import Any, AsyncIterator, Iterator, List, Optional
import uuid

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


GET_EXECUTORS_PAGINATED = """-- name: get_executors_paginated \\:many
SELECT 
    ROW_NUMBER() OVER (ORDER BY started DESC NULLS LAST) as row_number,
    id,
    name,
    worker_name,
    queue_pattern,
    queues,
    started,
    stopped,
    last_heartbeat,
    stats,
    status
FROM hyrex_executor
ORDER BY started DESC NULLS LAST
LIMIT :p1 OFFSET :p2
"""


@dataclasses.dataclass()
class GetExecutorsPaginatedParams:
    limit: int
    offset: int


@dataclasses.dataclass()
class GetExecutorsPaginatedRow:
    row_number: int
    id: uuid.UUID
    name: str
    worker_name: str
    queue_pattern: str
    queues: List[str]
    started: Optional[datetime.datetime]
    stopped: Optional[datetime.datetime]
    last_heartbeat: Optional[datetime.datetime]
    stats: Optional[Any]
    status: Any


class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def get_executors_paginated(self, arg: GetExecutorsPaginatedParams) -> Iterator[GetExecutorsPaginatedRow]:
        result = self._conn.execute(sqlalchemy.text(GET_EXECUTORS_PAGINATED), {"p1": arg.limit, "p2": arg.offset})
        for row in result:
            yield GetExecutorsPaginatedRow(
                row_number=row[0],
                id=row[1],
                name=row[2],
                worker_name=row[3],
                queue_pattern=row[4],
                queues=row[5],
                started=row[6],
                stopped=row[7],
                last_heartbeat=row[8],
                stats=row[9],
                status=row[10],
            )


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def get_executors_paginated(self, arg: GetExecutorsPaginatedParams) -> AsyncIterator[GetExecutorsPaginatedRow]:
        result = await self._conn.stream(sqlalchemy.text(GET_EXECUTORS_PAGINATED), {"p1": arg.limit, "p2": arg.offset})
        async for row in result:
            yield GetExecutorsPaginatedRow(
                row_number=row[0],
                id=row[1],
                name=row[2],
                worker_name=row[3],
                queue_pattern=row[4],
                queues=row[5],
                started=row[6],
                stopped=row[7],
                last_heartbeat=row[8],
                stats=row[9],
                status=row[10],
            )



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: _09_create_system_log_table.sql
import dataclasses

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


CREATE_SYSTEM_LOG_TABLE = """-- name: create_system_log_table \\:exec
CREATE TABLE IF NOT EXISTS hyrex_system_logs
(
    id         UUID    NOT NULL PRIMARY KEY,
    timestamp  TIMESTAMP WITH TIME ZONE,
    event_name VARCHAR NOT NULL,
    event_body JSON    NOT NULL
)
"""


@dataclasses.dataclass()
class CreateSystemLogTableParams:

    pass

class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def create_system_log_table(self, arg: CreateSystemLogTableParams) -> None:
        self._conn.execute(sqlalchemy.text(CREATE_SYSTEM_LOG_TABLE))


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def create_system_log_table(self, arg: CreateSystemLogTableParams) -> None:
        await self._conn.execute(sqlalchemy.text(CREATE_SYSTEM_LOG_TABLE))



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: transition_task_state.sql
import dataclasses
from typing import Any, Optional
import uuid

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


TRANSITION_TASK_STATE = """-- name: transition_task_state \\:one
SELECT transition_task_state(
    :p1\\:\\:UUID,
    :p2\\:\\:task_run_status
) AS task_run
"""


@dataclasses.dataclass()
class TransitionTaskStateParams:
    task_id: uuid.UUID
    next_state: Any


class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def transition_task_state(self, arg: TransitionTaskStateParams) -> Optional[Optional[Any]]:
        row = self._conn.execute(sqlalchemy.text(TRANSITION_TASK_STATE), {"p1": arg.task_id, "p2": arg.next_state}).first()
        if row is None:
            return None
        return row[0]


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def transition_task_state(self, arg: TransitionTaskStateParams) -> Optional[Optional[Any]]:
        row = (await self._conn.execute(sqlalchemy.text(TRANSITION_TASK_STATE), {"p1": arg.task_id, "p2": arg.next_state})).first()
        if row is None:
            return None
        return row[0]



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: get_project_stats.sql
import dataclasses
import datetime
from typing import AsyncIterator, Iterator, Optional

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


GET_PROJECT_STATS = """-- name: get_project_stats \\:many
SELECT 
    timepoint,
    queued,
    running,
    waiting,
    failed,
    success,
    lost,
    queued + running + waiting + failed + lost as total,
    queued_delta,
    success_delta,
    failed_delta,
    lost_delta
FROM hyrex_stats_task_status_counts
WHERE timepoint >= NOW() - make_interval(mins => :p1\\:\\:int)
ORDER BY timepoint DESC
"""


@dataclasses.dataclass()
class GetProjectStatsParams:
    minutes: int


@dataclasses.dataclass()
class GetProjectStatsRow:
    timepoint: datetime.datetime
    queued: Optional[int]
    running: Optional[int]
    waiting: Optional[int]
    failed: Optional[int]
    success: Optional[int]
    lost: Optional[int]
    total: int
    queued_delta: Optional[int]
    success_delta: Optional[int]
    failed_delta: Optional[int]
    lost_delta: Optional[int]


class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def get_project_stats(self, arg: GetProjectStatsParams) -> Iterator[GetProjectStatsRow]:
        result = self._conn.execute(sqlalchemy.text(GET_PROJECT_STATS), {"p1": arg.minutes})
        for row in result:
            yield GetProjectStatsRow(
                timepoint=row[0],
                queued=row[1],
                running=row[2],
                waiting=row[3],
                failed=row[4],
                success=row[5],
                lost=row[6],
                total=row[7],
                queued_delta=row[8],
                success_delta=row[9],
                failed_delta=row[10],
                lost_delta=row[11],
            )


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def get_project_stats(self, arg: GetProjectStatsParams) -> AsyncIterator[GetProjectStatsRow]:
        result = await self._conn.stream(sqlalchemy.text(GET_PROJECT_STATS), {"p1": arg.minutes})
        async for row in result:
            yield GetProjectStatsRow(
                timepoint=row[0],
                queued=row[1],
                running=row[2],
                waiting=row[3],
                failed=row[4],
                success=row[5],
                lost=row[6],
                total=row[7],
                queued_delta=row[8],
                success_delta=row[9],
                failed_delta=row[10],
                lost_delta=row[11],
            )



from sqlalchemy import create_engine

from hyrex.dispatcher.sqlc import (
    create_cron_job_for_sql_query_sync,
    create_cron_job_for_sql_query,
    create_enums_sync,
    create_tables_sync,
    create_functions_sync,
)


def init_postgres_db(conn_string):
    """Initialize the Postgres database with all required Hyrex tables and functions."""
    # Convert connection string to use psycopg3
    if conn_string.startswith("postgresql://"):
        conn_string = conn_string.replace("postgresql://", "postgresql+psycopg://", 1)
    elif conn_string.startswith("postgres://"):
        conn_string = conn_string.replace("postgres://", "postgresql+psycopg://", 1)

    # Create SQLAlchemy engine
    engine = create_engine(conn_string)

    with engine.begin() as conn:
        # Create enums (will skip if they already exist)
        try:
            create_enums_sync(conn)
        except Exception:
            pass  # Enums might already exist

        # Create tables and indexes
        create_tables_sync(conn)

        # Create functions and triggers
        create_functions_sync(conn)

        # TODO: Register cron jobs for system tasks



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: save_result.sql
import dataclasses
from typing import Any, Optional
import uuid

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


SAVE_RESULT = """-- name: save_result \\:exec
INSERT INTO hyrex_task_result
    (task_id, result)
VALUES (:p1, :p2)
"""


@dataclasses.dataclass()
class SaveResultParams:
    task_id: uuid.UUID
    result: Optional[Any]


class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def save_result(self, arg: SaveResultParams) -> None:
        self._conn.execute(sqlalchemy.text(SAVE_RESULT), {"p1": arg.task_id, "p2": arg.result})


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def save_result(self, arg: SaveResultParams) -> None:
        await self._conn.execute(sqlalchemy.text(SAVE_RESULT), {"p1": arg.task_id, "p2": arg.result})



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: pull_active_cron_expressions.sql
import dataclasses
import datetime
from typing import AsyncIterator, Iterator, Optional

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


PULL_ACTIVE_CRON_EXPRESSIONS = """-- name: pull_active_cron_expressions \\:many
SELECT jobid,
       schedule,
       command,
       active,
       jobname,
       activated_at,
       scheduled_jobs_confirmed_until,
       should_backfill
FROM hyrex_cron_job
WHERE active = true
"""


@dataclasses.dataclass()
class PullActiveCronExpressionsParams:

    pass

@dataclasses.dataclass()
class PullActiveCronExpressionsRow:
    jobid: int
    schedule: Optional[str]
    command: str
    active: bool
    jobname: str
    activated_at: Optional[datetime.datetime]
    scheduled_jobs_confirmed_until: Optional[datetime.datetime]
    should_backfill: Optional[bool]


class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def pull_active_cron_expressions(self, arg: PullActiveCronExpressionsParams) -> Iterator[PullActiveCronExpressionsRow]:
        result = self._conn.execute(sqlalchemy.text(PULL_ACTIVE_CRON_EXPRESSIONS))
        for row in result:
            yield PullActiveCronExpressionsRow(
                jobid=row[0],
                schedule=row[1],
                command=row[2],
                active=row[3],
                jobname=row[4],
                activated_at=row[5],
                scheduled_jobs_confirmed_until=row[6],
                should_backfill=row[7],
            )


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def pull_active_cron_expressions(self, arg: PullActiveCronExpressionsParams) -> AsyncIterator[PullActiveCronExpressionsRow]:
        result = await self._conn.stream(sqlalchemy.text(PULL_ACTIVE_CRON_EXPRESSIONS))
        async for row in result:
            yield PullActiveCronExpressionsRow(
                jobid=row[0],
                schedule=row[1],
                command=row[2],
                active=row[3],
                jobname=row[4],
                activated_at=row[5],
                scheduled_jobs_confirmed_until=row[6],
                should_backfill=row[7],
            )



import collections
import json
from typing import Sequence

from hyrex.task_wrapper import TaskWrapper
from hyrex.workflow.workflow_builder_context import (
    clear_current_workflow_builder, set_current_workflow_builder)


class DagNode:
    def __init__(self, task_wrapper: TaskWrapper, workflow_builder: "WorkflowBuilder"):
        self.task_wrapper = task_wrapper
        self.workflow_builder = workflow_builder
        self.children = []
        self.parents = []

    def has_path_to(self, target: "DagNode", visited=None) -> bool:
        if visited is None:
            visited = set()
        if self is target:
            return True
        visited.add(self)
        for child in self.children:
            if child not in visited and child.has_path_to(target, visited):
                return True
        return False

    def add_child(self, child: "DagNode"):
        if child.has_path_to(self):
            raise ValueError("Adding this child would create a cycle!")
        if child in self.children:
            raise ValueError("This node is already a child of mine!")

        self.children.append(child)
        child.parents.append(self)

    def get_children(self) -> list["DagNode"]:
        return self.children

    def __rshift__(self, other: TaskWrapper | Sequence[TaskWrapper] | "DagNode"):
        if isinstance(other, TaskWrapper):
            node = self.workflow_builder.get_or_create_node(other)
            self.add_child(node)
            return node
        elif isinstance(other, collections.abc.Sequence):
            nodes = [
                (
                    self.workflow_builder.get_or_create_node(task)
                    if isinstance(task, TaskWrapper)
                    else task
                )
                for task in other
            ]
            for node in nodes:
                self.add_child(node)
            return nodes
        elif isinstance(other, DagNode):
            self.add_child(other)
            return other
        else:
            raise TypeError(f"Unknown type of {other}, {type(other)}")

    def __rrshift__(self, other: Sequence["TaskWrapper | DagNode"]) -> "DagNode":
        if not isinstance(other, collections.abc.Sequence):
            raise TypeError(f"Unknown type of {other}, {type(other)}")

        for item in other:
            if isinstance(item, TaskWrapper):
                node = self.workflow_builder.get_or_create_node(item)
                node.add_child(self)
            elif isinstance(item, DagNode):
                item.add_child(self)
            else:
                raise TypeError(
                    f"Cannot use object {item} of type {type(item)} as node in DAG"
                )

        return self  # Return self to allow chaining like task1 >> task2 >> task3


class WorkflowBuilder:
    def __init__(self):
        self.nodes = {}  # Map from TaskWrapper to DagNode

    def __enter__(self) -> "WorkflowBuilder":
        set_current_workflow_builder(self)
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        clear_current_workflow_builder()

    def get_or_create_node(self, task_wrapper: TaskWrapper) -> DagNode:
        if task_wrapper not in self.nodes:
            self.nodes[task_wrapper] = DagNode(
                task_wrapper=task_wrapper, workflow_builder=self
            )
        return self.nodes[task_wrapper]

    def get_root_nodes(self):
        return [node for node in self.nodes.values() if not node.parents]

    def to_json(self) -> dict:
        """
        Returns a JSON dict representing the workflow DAG.
        The JSON has the format:
        {
          "nodes": [
            { "id": "1", "name": "TaskIdentifier1" },
            { "id": "2", "name": "TaskIdentifier2" },
            ...
          ],
          "edges": [
            { "from": "1", "to": "2" },
            { "from": "1", "to": "3" },
            ...
          ]
        }
        """
        nodes_dict = {}
        edges = []
        visited = set()
        node_ids = {}  # Maps node (via id(node)) to a unique increasing integer.
        next_id = [1]  # Using a list to allow modification in nested scope.

        def dfs(node: DagNode):
            node_key = id(node)
            if node_key in visited:
                return
            visited.add(node_key)

            # Assign a unique increasing id to the node if not already assigned.
            if node_key not in node_ids:
                node_ids[node_key] = next_id[0]
                next_id[0] += 1

            # Use task_wrapper.task_identifier as the node's name.
            node_name = getattr(
                node.task_wrapper, "task_identifier", f"node_{node_ids[node_key]}"
            )
            nodes_dict[node_ids[node_key]] = {
                "id": str(node_ids[node_key]),
                "name": node_name,
            }

            for child in node.children:
                child_key = id(child)
                if child_key not in node_ids:
                    node_ids[child_key] = next_id[0]
                    next_id[0] += 1
                edges.append(
                    {"from": str(node_ids[node_key]), "to": str(node_ids[child_key])}
                )
                dfs(child)

        for root in self.get_root_nodes():
            dfs(root)

        # TODO: Consider using a more specific schema
        return {"nodes": list(nodes_dict.values()), "edges": edges}



from enum import StrEnum

from pydantic import BaseModel, Field, field_validator

from hyrex.hyrex_queue import HyrexQueue


class ConfigPhase(StrEnum):
    # Config set on HyrexRegistry
    registry = "registry"
    # Config set on decorator
    decorator = "decorator"
    # Config set at send time
    send = "send"


class WorkflowConfig(BaseModel):
    config_phase: ConfigPhase

    queue: str | HyrexQueue | None = None  # Overrides task queue
    priority: int | None = Field(default=None, ge=1, le=10)  # Overrides task priority

    timeout_seconds: int | None = Field(default=None, gt=0)  # For workflow
    idempotency_key: str | None = None  # For workflow

    @field_validator("config_phase")
    @classmethod  # This is optional in Pydantic v2
    def validate_username(cls, value):
        if value == ConfigPhase.registry:
            raise ValueError(
                "WorkflowConfigs should not be defined at the registry level."
            )
        return value

    def get_default_config(self):
        assert (
            self.config_phase == ConfigPhase.decorator
        ), "Default config must be specified at decorator-time."
        default_config = {}
        if self.queue:
            default_config["queue"] = self.get_queue_name()
        if self.priority:
            default_config["priority"] = self.priority
        if self.timeout_seconds:
            default_config["timeout_seconds"] = self.timeout_seconds
        return default_config

    def get_queue_name(self) -> str:
        if isinstance(self.queue, str):
            return self.queue
        else:
            return self.queue.name

    def merge(self, other: "WorkflowConfig") -> "WorkflowConfig":
        """Merge another config into this one, with other taking precedence"""
        if self.config_phase == ConfigPhase.send:
            assert other.config_phase != ConfigPhase.decorator

        merged_data = self.model_dump()
        other_data = other.model_dump()

        for field, value in other_data.items():
            if value is not None:
                merged_data[field] = value

        return WorkflowConfig(**merged_data)


class TaskConfig(BaseModel):
    config_phase: ConfigPhase

    queue: str | HyrexQueue | None = None
    priority: int | None = Field(default=None, ge=1, le=10)
    max_retries: int | None = Field(default=None, ge=0)
    timeout_seconds: int | None = Field(default=None, gt=0)
    idempotency_key: str | None = None

    def get_default_config(self):
        assert (
            self.config_phase == ConfigPhase.decorator
        ), "Default config must be specified at decorator-time."
        default_config = {}
        if self.queue:
            default_config["queue"] = self.get_queue_name()
        if self.priority:
            default_config["priority"] = self.priority
        if self.max_retries:
            default_config["max_retries"] = self.max_retries
        if self.timeout_seconds:
            default_config["timeout_seconds"] = self.timeout_seconds
        return default_config

    def get_queue_name(self) -> str:
        if isinstance(self.queue, str):
            return self.queue
        else:
            return self.queue.name

    def merge(self, other: "TaskConfig") -> "TaskConfig":
        """Merge another config into this one, with other taking precedence"""
        if self.config_phase == ConfigPhase.decorator:
            assert other.config_phase != ConfigPhase.registry
        elif self.config_phase == ConfigPhase.send:
            assert other.config_phase not in [
                ConfigPhase.decorator,
                ConfigPhase.registry,
            ]

        merged_data = self.model_dump()
        other_data = other.model_dump()

        for field, value in other_data.items():
            if value is not None:
                merged_data[field] = value

        return TaskConfig(**merged_data)

    def apply_workflow_config(self, workflow_config: WorkflowConfig) -> "TaskConfig":
        """Apply workflow config settings to this task config"""
        if workflow_config is None:
            return self

        updated_config = self.model_copy()

        if workflow_config.queue is not None:
            updated_config.queue = workflow_config.queue
        if workflow_config.priority is not None:
            updated_config.priority = workflow_config.priority

        return updated_config



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: update_cron_job_confirmation_ts.sql
import dataclasses

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


UPDATE_CRON_JOB_CONFIRMATION_TS = """-- name: update_cron_job_confirmation_ts \\:exec
UPDATE hyrex_cron_job
SET scheduled_jobs_confirmed_until = now()
WHERE jobid = :p1
"""


@dataclasses.dataclass()
class UpdateCronJobConfirmationTsParams:
    jobid: int


class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def update_cron_job_confirmation_ts(self, arg: UpdateCronJobConfirmationTsParams) -> None:
        self._conn.execute(sqlalchemy.text(UPDATE_CRON_JOB_CONFIRMATION_TS), {"p1": arg.jobid})


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def update_cron_job_confirmation_ts(self, arg: UpdateCronJobConfirmationTsParams) -> None:
        await self._conn.execute(sqlalchemy.text(UPDATE_CRON_JOB_CONFIRMATION_TS), {"p1": arg.jobid})



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: create_task_run.sql
import dataclasses
import datetime
from typing import Any, List, Optional
import uuid

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


CREATE_TASK_RUN = """-- name: create_task_run \\:one
SELECT create_task_run(
    :p1\\:\\:UUID,
    :p2\\:\\:UUID,
    :p3\\:\\:UUID,
    :p4\\:\\:UUID,
    :p5\\:\\:task_run_status,
    :p6\\:\\:VARCHAR,
    :p7\\:\\:JSON,
    :p8\\:\\:VARCHAR,
    :p9\\:\\:SMALLINT,
    :p10\\:\\:SMALLINT,
    :p11\\:\\:INT,
    :p12\\:\\:VARCHAR,
    :p13\\:\\:TIMESTAMP WITH TIME ZONE,
    :p14\\:\\:UUID,
    :p15\\:\\:UUID[]
) AS id
"""


@dataclasses.dataclass()
class CreateTaskRunParams:
    id: uuid.UUID
    durable_id: uuid.UUID
    root_id: uuid.UUID
    parent_id: uuid.UUID
    status: Any
    task_name: str
    args: Any
    queue: str
    max_retries: int
    priority: int
    timeout_seconds: int
    idempotency_key: str
    scheduled_start: datetime.datetime
    workflow_run_id: uuid.UUID
    workflow_dependencies: List[uuid.UUID]


CREATE_TASK_RUN_FUNCTION = """-- name: create_task_run_function \\:exec
CREATE OR REPLACE FUNCTION create_task_run(
    p_id UUID,
    p_durable_id UUID,
    p_root_id UUID,
    p_parent_id UUID,
    p_status task_run_status,
    p_task_name VARCHAR,
    p_args JSON,
    p_queue VARCHAR,
    p_max_retries SMALLINT,
    p_priority SMALLINT,
    p_timeout_seconds INT DEFAULT NULL,
    p_idempotency_key VARCHAR DEFAULT NULL,
    p_scheduled_start TIMESTAMP WITH TIME ZONE DEFAULT NULL,
    p_workflow_run_id UUID DEFAULT NULL,
    p_workflow_dependencies UUID[] DEFAULT NULL
) RETURNS UUID AS $$
DECLARE
    v_task_id UUID;
    v_queued_timestamp TIMESTAMP WITH TIME ZONE;
BEGIN
    -- Validate inputs
    IF p_status = 'AWAIT_START_TIME' AND p_scheduled_start IS NULL THEN
        RAISE EXCEPTION 'scheduled_start is required when status is AWAIT_START_TIME';
    END IF;
    
    IF p_status NOT IN ('QUEUED', 'AWAIT_START_TIME', 'AWAIT_DEPS') THEN
        RAISE EXCEPTION 'Initial task status must be QUEUED, AWAIT_START_TIME, or AWAIT_DEPS';
    END IF;
    
    -- Set queued timestamp only for QUEUED status
    IF p_status = 'QUEUED' THEN
        v_queued_timestamp \\:= CURRENT_TIMESTAMP;
    ELSE
        v_queued_timestamp \\:= NULL;
    END IF;
    
    -- Insert the task
    INSERT INTO hyrex_task_run (
        id,
        durable_id,
        root_id,
        parent_id,
        status,
        task_name,
        args,
        queue,
        max_retries,
        priority,
        timeout_seconds,
        attempt_number,
        queued,
        idempotency_key,
        scheduled_start,
        workflow_run_id,
        workflow_dependencies
    )
    VALUES (
        p_id,
        p_durable_id,
        p_root_id,
        p_parent_id,
        p_status,
        p_task_name,
        p_args,
        p_queue,
        p_max_retries,
        p_priority,
        p_timeout_seconds,
        0,
        v_queued_timestamp,
        p_idempotency_key,
        p_scheduled_start,
        p_workflow_run_id,
        p_workflow_dependencies
    )
    ON CONFLICT (task_name, idempotency_key)
        WHERE idempotency_key IS NOT NULL
        DO NOTHING
    RETURNING id INTO v_task_id;
    
    -- Log idempotency conflict if it occurred
    IF v_task_id IS NULL AND p_idempotency_key IS NOT NULL THEN
        INSERT INTO hyrex_system_logs (id, timestamp, event_name, event_body)
        VALUES (
            gen_random_uuid(),
            CURRENT_TIMESTAMP,
            'task_idempotency_conflict',
            json_build_object(
                'task_name', p_task_name,
                'idempotency_key', p_idempotency_key,
                'attempted_task_id', p_id,
                'queue', p_queue,
                'status', p_status\\:\\:TEXT
            )
        );
    END IF;
    
    -- Return NULL if conflict occurred, otherwise return the new task ID
    RETURN v_task_id;
END;
$$ LANGUAGE plpgsql
"""


@dataclasses.dataclass()
class CreateTaskRunFunctionParams:

    pass

class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def create_task_run(self, arg: CreateTaskRunParams) -> Optional[uuid.UUID]:
        row = self._conn.execute(sqlalchemy.text(CREATE_TASK_RUN), {
            "p1": arg.id,
            "p2": arg.durable_id,
            "p3": arg.root_id,
            "p4": arg.parent_id,
            "p5": arg.status,
            "p6": arg.task_name,
            "p7": arg.args,
            "p8": arg.queue,
            "p9": arg.max_retries,
            "p10": arg.priority,
            "p11": arg.timeout_seconds,
            "p12": arg.idempotency_key,
            "p13": arg.scheduled_start,
            "p14": arg.workflow_run_id,
            "p15": arg.workflow_dependencies,
        }).first()
        if row is None:
            return None
        return row[0]

    def create_task_run_function(self, arg: CreateTaskRunFunctionParams) -> None:
        self._conn.execute(sqlalchemy.text(CREATE_TASK_RUN_FUNCTION))


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def create_task_run(self, arg: CreateTaskRunParams) -> Optional[uuid.UUID]:
        row = (await self._conn.execute(sqlalchemy.text(CREATE_TASK_RUN), {
            "p1": arg.id,
            "p2": arg.durable_id,
            "p3": arg.root_id,
            "p4": arg.parent_id,
            "p5": arg.status,
            "p6": arg.task_name,
            "p7": arg.args,
            "p8": arg.queue,
            "p9": arg.max_retries,
            "p10": arg.priority,
            "p11": arg.timeout_seconds,
            "p12": arg.idempotency_key,
            "p13": arg.scheduled_start,
            "p14": arg.workflow_run_id,
            "p15": arg.workflow_dependencies,
        })).first()
        if row is None:
            return None
        return row[0]

    async def create_task_run_function(self, arg: CreateTaskRunFunctionParams) -> None:
        await self._conn.execute(sqlalchemy.text(CREATE_TASK_RUN_FUNCTION))



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: _02_create_hyrex_app_table.sql
import dataclasses

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


CREATE_APP_TABLE = """-- name: create_app_table \\:exec
CREATE TABLE IF NOT EXISTS hyrex_app
(
    id       BIGSERIAL NOT NULL PRIMARY KEY,
    app_info JSON
)
"""


@dataclasses.dataclass()
class CreateAppTableParams:

    pass

class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def create_app_table(self, arg: CreateAppTableParams) -> None:
        self._conn.execute(sqlalchemy.text(CREATE_APP_TABLE))


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def create_app_table(self, arg: CreateAppTableParams) -> None:
        await self._conn.execute(sqlalchemy.text(CREATE_APP_TABLE))



import json
from typing import Any, Dict, List

from hyrex.schemas import CronJobRun, EnqueueTaskRequest

CREATE_HYREX_CRON_JOB_TABLE = """
DO $$
BEGIN
    IF NOT EXISTS (SELECT 1 FROM pg_type WHERE typname = 'job_source_type') THEN
        CREATE TYPE job_source_type AS ENUM ('SYSTEM', 'TASK', 'APPLICATION');
    END IF;
END
$$;

CREATE TABLE IF NOT EXISTS hyrex_cron_job
(
    jobid                          bigserial PRIMARY KEY,
    schedule                       text,
    command                        text    NOT NULL,
    active                         boolean NOT NULL DEFAULT true,
    jobname                        text    NOT NULL,
    job_source                     job_source_type NOT NULL,
    activated_at                   timestamptz      default now(),
    scheduled_jobs_confirmed_until timestamptz      default now(),
    should_backfill                boolean default true,
    UNIQUE (jobname)
);
"""

# Specifically modeled on cron.job_run_details table in pg_cron
CREATE_HYREX_CRON_JOB_RUN_DETAILS_TABLE = """
DO $$
BEGIN
    IF NOT EXISTS (SELECT 1 FROM pg_type WHERE typname = 'cron_job_status' AND typnamespace = 'public'::regnamespace) THEN
        CREATE TYPE public.cron_job_status AS ENUM (
            'success',
            'queued',
            'failed'
        );
    END IF;
END $$;

CREATE TABLE IF NOT EXISTS hyrex_cron_job_run_details (
  jobid        bigint      NOT NULL,
  runid        bigserial   PRIMARY KEY,
  command      text        NOT NULL,
  status       cron_job_status,
  schedule_time timestamptz not null,
  start_time   timestamptz,
  end_time     timestamptz,
  UNIQUE (jobid, schedule_time)
)
"""

CREATE_HYREX_SCHEDULER_LOCK_TABLE = """
    CREATE TABLE IF NOT EXISTS hyrex_scheduler_lock
    (
        lockid       bigserial PRIMARY KEY,
        worker_name  text        NOT NULL,
        acquired_at  timestamptz NOT NULL DEFAULT now(),
        heartbeat_at timestamptz NOT NULL DEFAULT now(),
        release_at   timestamptz NOT NULL,
        is_active    boolean     NOT NULL DEFAULT true
    );
"""

ACQUIRE_SCHEDULER_LOCK = """
INSERT INTO hyrex_scheduler_lock (
    lockid, worker_name, acquired_at, heartbeat_at, release_at, is_active
)
VALUES (
    1,                  -- single global lock id
    $1,                 -- workerName
    now(),              -- acquired_at
    now(),              -- heartbeat_at
    now() + CAST($2 AS interval),  -- release_at (e.g. now + '5 minutes')
    true                -- is_active
)
ON CONFLICT (lockid)
  DO UPDATE 
     SET worker_name  = EXCLUDED.worker_name,
         acquired_at  = EXCLUDED.acquired_at,
         heartbeat_at = EXCLUDED.heartbeat_at,
         release_at   = EXCLUDED.release_at,
         is_active    = EXCLUDED.is_active
   WHERE (
       -- Only update if the lock is not truly active,
       -- i.e. is already inactive or expired:
       hyrex_scheduler_lock.is_active = false
       OR hyrex_scheduler_lock.release_at <= now()
   )
RETURNING lockid;
"""

RELEASE_SCHEDULER_LOCK = """
    UPDATE hyrex_scheduler_lock
    SET is_active    = false,
        release_at   = now(),
        heartbeat_at = now()
    WHERE lockid = 1
      AND worker_name = $1
    RETURNING lockid;
"""

PULL_ACTIVE_CRON_EXPRESSIONS = """
    SELECT jobid,
           schedule,
           command,
           active,
           jobname,
           activated_at,
           scheduled_jobs_confirmed_until,
           should_backfill
    FROM hyrex_cron_job
    WHERE active = true
--       AND activated_at > NOW();
"""

UPDATE_CRON_JOB_CONFIRMATION_TS = """
    UPDATE hyrex_cron_job
    SET scheduled_jobs_confirmed_until = now()
    WHERE jobid = $1;
"""


def cron_job_runs_to_sql(runs: List[CronJobRun]) -> Dict[str, Any]:
    """
    Convert a list of cron job runs to SQL insert statement with dollar-sign parameterized values.

    Args:
        runs: List of CronJobRun objects with jobid, command, and schedule_time attributes

    Returns:
        dict: Contains 'sql' string and 'values' list for parameterized query
    """
    # Calculate the placeholder indices for each run
    placeholder_index = 1
    placeholder_strings = []

    # Collect all values in order
    values = []
    for run in runs:
        values.extend([run.jobid, run.command, "queued", run.schedule_time.isoformat()])

    for _ in runs:
        # Use $1, $2, etc. format instead of %(1)s
        placeholder_strings.append(
            f"(${placeholder_index}, ${placeholder_index + 1}, ${placeholder_index + 2}, ${placeholder_index + 3})"
        )
        placeholder_index += 4

    # Build this part separately to avoid Python3.11 f-string backslash limitations
    placeholder_section = ",\n            ".join(placeholder_strings)

    # Construct the final SQL string
    sql = f"""
        INSERT INTO hyrex_cron_job_run_details
            (jobid, command, status, schedule_time)
        VALUES
            {placeholder_section}
        ON CONFLICT (jobid, schedule_time) DO NOTHING
        RETURNING runid;
    """

    return {"sql": sql, "values": values}


CREATE_EXECUTE_QUEUED_COMMAND_FUNCTION = """
CREATE OR REPLACE FUNCTION execute_queued_command()
RETURNS text AS
$$
DECLARE
    cmd text;
    selected_runid bigint;
    start_ts timestamp;
BEGIN
    -- 1) Grab the first queued command (row) and lock it
    SELECT command, runid 
      INTO cmd, selected_runid
      FROM hyrex_cron_job_run_details
     WHERE status = 'queued'
       AND schedule_time <= now()
     ORDER BY schedule_time
     LIMIT 1
     FOR UPDATE SKIP LOCKED;

    IF FOUND THEN
        -- 2) Mark start_time right before execution
        start_ts := clock_timestamp();

        -- 3) Execute the command text
        EXECUTE cmd;

        -- 4) Mark the end_time and set status
        UPDATE hyrex_cron_job_run_details
           SET status      = 'success',
               start_time  = start_ts,
               end_time    = clock_timestamp()
         WHERE runid = selected_runid
           AND status = 'queued';
        
        RETURN 'executed';
    END IF;

    RETURN 'not_found';

EXCEPTION
    WHEN OTHERS THEN
        -- In the event of an error, mark job as failed and record times
        UPDATE hyrex_cron_job_run_details
           SET status      = 'failed',
               start_time  = COALESCE(start_ts, clock_timestamp()),
               end_time    = clock_timestamp()
         WHERE runid = selected_runid
           AND status = 'queued';
        RAISE;
END;
$$
LANGUAGE plpgsql;
"""


def create_insert_task_cron_expression(enqueue_task_request: EnqueueTaskRequest) -> str:
    tr = enqueue_task_request

    # Handle idempotency key exactly like the TypeScript version
    idempotency_key_str = (
        "NULL" if tr.idempotency_key is None else f"'{tr.idempotency_key}'"
    )

    # Convert args to JSON string
    args_json = json.dumps(tr.args)

    # Handle timeout_seconds which could be None
    timeout_seconds_str = (
        "NULL" if tr.timeout_seconds is None else str(tr.timeout_seconds)
    )

    sql = f"""WITH vars AS (
       SELECT gen_random_uuid() as shared_uuid
   ),
   task_insertion AS (
   INSERT INTO hyrex_task_run (
                                     id,
                                     durable_id,
                                     root_id,
                                     task_name,
                                     args,
                                     queue,
                                     max_retries,
                                     priority,
                                     timeout_seconds,
                                     status,
                                     attempt_number,
                                     queued,
                                     idempotency_key
       )
       SELECT
               v.shared_uuid,
               v.shared_uuid,
               v.shared_uuid,
               '{tr.task_name}',
               '{args_json}'::json,
               '{tr.queue}',
               {tr.max_retries},
               {tr.priority},
               {timeout_seconds_str},
               'queued'::task_run_status,
               0,
               CURRENT_TIMESTAMP,
               {idempotency_key_str}
       FROM vars v
       ON CONFLICT (task_name, idempotency_key)
           WHERE idempotency_key IS NOT NULL
           DO NOTHING
       RETURNING id),
    log_entry AS (
        INSERT INTO hyrex_system_logs (
                                       id,
                                       timestamp,
                                       event_name,
                                       event_body
            )
            SELECT gen_random_uuid(),
                   CURRENT_TIMESTAMP,
                   'IDEMPOTENCY_COLLISION',
                   json_build_object(
                           'attempted_task_id', v.shared_uuid,
                           'idempotency_key', {idempotency_key_str},
                           'task_name', '{tr.task_name}',
                           'queue', '{tr.queue}'
                   )
            FROM vars v
            WHERE NOT EXISTS (SELECT 1 FROM task_insertion)
              AND {idempotency_key_str} IS NOT NULL)
SELECT (SELECT id FROM task_insertion) as task_created;"""

    return sql


CREATE_CRON_JOB_FOR_TASK = """
    INSERT INTO hyrex_cron_job (schedule, command, jobname, job_source)
    VALUES ($1, $2, $3, 'TASK')
    ON CONFLICT (jobname) 
    DO UPDATE SET 
        activated_at = CURRENT_TIMESTAMP,
        schedule = EXCLUDED.schedule,
        command = EXCLUDED.command,
        job_source = 'TASK',
        active = true;
"""

CREATE_CRON_JOB_FOR_SQL_QUERY = """
    INSERT INTO hyrex_cron_job (schedule, command, jobname, should_backfill, job_source)
    VALUES ($1, $2, $3, $4, 'SYSTEM')
    ON CONFLICT (jobname) 
    DO UPDATE SET 
        activated_at = CURRENT_TIMESTAMP,
        schedule = EXCLUDED.schedule,
        command = EXCLUDED.command,
        should_backfill = EXCLUDED.should_backfill,
        job_source = 'SYSTEM',
        active = true;
"""

TURN_OFF_CRON_FOR_TASK = """
    UPDATE hyrex_cron_job
    SET active   = false,
        schedule = NULL
    WHERE jobname = $1
      AND job_source = 'TASK';
"""



import atexit
import logging
import signal
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Type
from uuid import UUID

from pydantic import BaseModel

from hyrex import constants
from hyrex.configs import TaskConfig
from hyrex.hyrex_queue import HyrexQueue
from hyrex.schemas import (
    CronJob,
    CronJobRun,
    DequeuedTask,
    EnqueueTaskRequest,
    QueuePattern,
    TaskRun,
    TaskStatus,
    WorkflowRunRequest,
)


class Dispatcher(ABC):
    logger = logging.getLogger(__name__)

    def _signal_handler(self, signum, frame):
        signame = signal.Signals(signum).name
        self.logger.debug(f"\nReceived {signame}. Shutting down Hyrex dispatcher...")
        self.stop()

    def _chain_signal_handlers(self, new_handler, old_handler):
        """Return a function that calls both the new and old signal handlers."""

        def wrapper(signum, frame):
            # Call the new handler first
            new_handler(signum, frame)
            # Then call the previous handler (if it exists)
            if old_handler and callable(old_handler):
                old_handler(signum, frame)

        return wrapper

    def _setup_signal_handlers(self):
        for sig in (signal.SIGTERM, signal.SIGINT):
            old_handler = signal.getsignal(sig)  # Get the existing handler
            new_handler = self._signal_handler  # Your new handler
            # Set the new handler, which calls both new and old handlers
            signal.signal(sig, self._chain_signal_handlers(new_handler, old_handler))

    def register_shutdown_handlers(self):
        self._setup_signal_handlers()
        atexit.register(self.stop)

    def __init__(self):
        pass

    @abstractmethod
    def register_app(self, app_info: dict):
        pass

    @abstractmethod
    def enqueue(
        self,
        tasks: list[EnqueueTaskRequest],
    ):
        pass

    @abstractmethod
    def dequeue(
        self,
        executor_id: UUID,
        task_names: list[str],
        queue: str = constants.ANY_QUEUE,
        concurrency_limit: int = 0,
    ) -> DequeuedTask:
        pass

    # Result must be a JSON string)
    @abstractmethod
    def mark_success(self, task_id: UUID, result: str):
        pass

    @abstractmethod
    def mark_failed(self, task_id: UUID):
        pass

    @abstractmethod
    def set_log_link(self, task_id: UUID, log_link: str):
        pass

    @abstractmethod
    def retry_task(self, task_id: UUID, backoff_seconds: int):
        pass

    @abstractmethod
    def try_to_cancel_task(self, task_id: UUID):
        pass

    @abstractmethod
    def try_to_cancel_durable_run(self, durable_id: UUID):
        pass

    @abstractmethod
    def task_canceled(self, task_id: UUID):
        pass

    # TODO: Remove?
    @abstractmethod
    def get_result(self, task_id: UUID) -> dict:
        pass

    @abstractmethod
    def get_task_status(self, task_id: UUID) -> TaskStatus:
        pass

    @abstractmethod
    def register_task_def(
        self,
        task_name: str,
        arg_schema: Type[BaseModel] | None,
        task_config: TaskConfig,
        cron: str = None,
        source_code: str = None,
    ):
        pass

    @abstractmethod
    def register_executor(
        self,
        executor_id: UUID,
        executor_name: str,
        queue_pattern: str,
        queues: list[HyrexQueue],
        worker_name: str,
    ):
        pass

    @abstractmethod
    def disconnect_executor(self, executor_id: UUID):
        pass

    @abstractmethod
    def executor_heartbeat(self, executor_ids: list[UUID], timestamp: datetime):
        pass

    @abstractmethod
    def update_executor_stats(self, executor_id: UUID, stats: dict):
        pass

    @abstractmethod
    def task_heartbeat(self, task_ids: list[UUID], timestamp: datetime):
        pass

    @abstractmethod
    def get_tasks_up_for_cancel(self) -> list[UUID]:
        pass

    @abstractmethod
    def get_queues_for_pattern(self, pattern: QueuePattern) -> list[str]:
        pass

    @abstractmethod
    def acquire_scheduler_lock(self, worker_name: str) -> int | None:
        pass

    @abstractmethod
    def pull_cron_job_expressions(self) -> list[CronJob]:
        pass

    @abstractmethod
    def update_cron_job_confirmation_timestamp(self, jobid: UUID):
        pass

    @abstractmethod
    def schedule_cron_job_runs(self, cron_job_runs: list[CronJobRun]):
        pass

    @abstractmethod
    def execute_queued_cron_job_run(self) -> str | None:
        pass

    @abstractmethod
    def release_scheduler_lock(self, worker_name: str) -> None:
        pass

    @abstractmethod
    def stop(self):
        pass

    @abstractmethod
    def register_workflow(
        self,
        name: str,
        source_code: str,
        workflow_dag_json: str,
        workflow_arg_schema: Type[BaseModel] | None,
        default_config: dict,
        cron: str | None,
    ):
        pass

    @abstractmethod
    def send_workflow_run(self, workflow_run_request: WorkflowRunRequest) -> UUID:
        pass

    @abstractmethod
    def advance_workflow_run(self, workflow_run_id: UUID):
        pass

    @abstractmethod
    def get_workflow_run_args(self, workflow_run_id: UUID) -> dict:
        pass

    @abstractmethod
    def register_cron_sql_query(
        self,
        cron_job_name: str,
        cron_sql_query: str,
        cron_expr: str,
        should_backfill: bool,
    ) -> None:
        pass

    @abstractmethod
    def get_durable_run_tasks(self, durable_id: UUID) -> list[TaskRun]:
        pass

    @abstractmethod
    def get_workflow_durable_runs(self, workflow_run_id: UUID) -> list[UUID]:
        pass

    @abstractmethod
    def update_executor_queues(self, executor_id: UUID, queues: list[str]):
        pass

    @abstractmethod
    def kv_set(self, key: str, value: str) -> None:
        pass

    @abstractmethod
    def kv_get(self, key: str) -> str | None:
        pass

    @abstractmethod
    def kv_delete(self, key: str) -> None:
        pass



import asyncio
import logging
import re
import time
from inspect import signature
from typing import Any, Callable, Generic, ParamSpec, TypeVar, get_type_hints, overload

import psycopg
from pydantic import BaseModel, ValidationError
from uuid6 import uuid7

from hyrex.configs import ConfigPhase, TaskConfig
from hyrex.dispatcher import Dispatcher
from hyrex.durable_run import DurableTaskRun
from hyrex.hyrex_context import get_hyrex_context
from hyrex.hyrex_queue import HyrexQueue
from hyrex.schemas import EnqueueTaskRequest, TaskStatus
from hyrex.workflow.workflow_builder_context import get_current_workflow_builder


def validate_error_handler(handler: Callable) -> None:
    sig = signature(handler)
    params = sig.parameters

    if len(params) > 1:
        raise ValueError("Hyrex on_error handler must accept either 0 or 1 arguments")

    if len(params) == 1:
        # Get the first (and only) parameter
        param = next(iter(params.values()))
        print(param.annotation)
        # Check its type annotation
        if param.annotation == param.empty:
            raise ValueError("Hyrex on_error handler must have type annotated args")
        if not issubclass(param.annotation, Exception):
            raise ValueError(
                "Hyrex on_error handler argument must be of type Exception"
            )


P = ParamSpec("P")  # Captures the parameter specification of the wrapped function
R = TypeVar("R")  # Captures the return type of the wrapped function


class TaskWrapper(Generic[P, R]):
    class ParamInfo(BaseModel):
        """Pydantic model to store parameter information"""

        type_hint: Any
        default: Any = None
        has_default: bool = False

        def __repr__(self):
            type_name = (
                self.type_hint.__name__
                if hasattr(self.type_hint, "__name__")
                else str(self.type_hint)
            )
            return f"ParamInfo(type_hint={type_name}, default={self.default}, has_default={self.has_default})"

        class Config:
            arbitrary_types_allowed = True  # Allow any Python type in type_hint field

    def __init__(
        self,
        task_identifier: str,
        func: Callable[P, R],
        dispatcher: Dispatcher,
        cron: str | None,
        task_config: TaskConfig,
        on_error: Callable = None,
        retry_backoff: int | Callable[[int], int] | None = None,
    ):
        self.logger = logging.getLogger(__name__)

        self.task_identifier = task_identifier
        self.func = func
        self.signature = signature(func)
        self.type_hints = get_type_hints(func)

        # TODO: Validate cron
        self.cron = cron

        self.task_config = task_config

        self.dispatcher = dispatcher
        self.on_error = on_error
        self.retry_backoff = retry_backoff

        if self.on_error:
            validate_error_handler(self.on_error)

        # Track the parameter names and their type hints for validation
        # Enforce that all parameters have type hints
        self.param_info = {}
        for param_name, param in self.signature.parameters.items():
            # Check if type hint exists for this parameter
            if param_name not in self.type_hints:
                raise TypeError(
                    f"Hyrex expects all task arguments to have type hints. Argument '{param_name}' in task '{task_identifier}' has no type hint."
                )

            # Create a Pydantic model instance for this parameter
            self.param_info[param_name] = self.ParamInfo(
                type_hint=self.type_hints.get(param_name),
                default=param.default if param.default is not param.empty else None,
                has_default=param.default is not param.empty,
            )

    async def async_call(self, **kwargs):
        self.logger.info(f"Executing task {self.func.__name__}.")

        # Validate kwargs against expected parameters
        validated_kwargs = self._validate_kwargs(kwargs)

        if asyncio.iscoroutinefunction(self.func):
            return await self.func(**validated_kwargs)
        else:
            return self.func(**validated_kwargs)

    def with_config(
        self,
        queue: str | HyrexQueue = None,
        priority: int = None,
        max_retries: int = None,
        timeout_seconds: int = None,
        idempotency_key: str = None,
    ) -> "TaskWrapper":
        new_task_config = TaskConfig(
            config_phase=ConfigPhase.send,
            queue=queue,
            priority=priority,
            max_retries=max_retries,
            timeout_seconds=timeout_seconds,
            idempotency_key=idempotency_key,
        )
        new_wrapper = TaskWrapper(
            task_identifier=self.task_identifier,
            func=self.func,
            dispatcher=self.dispatcher,
            cron=self.cron,
            task_config=self.task_config.merge(new_task_config),
            on_error=self.on_error,
            retry_backoff=self.retry_backoff,
        )
        return new_wrapper

    def get_queue(self) -> HyrexQueue | str:
        return self.task_config.queue

    def get_retry_backoff(self, attempt_number: int) -> int:
        if self.retry_backoff is None:
            return 0
        elif isinstance(self.retry_backoff, int):
            return self.retry_backoff
        elif callable(self.retry_backoff):
            return self.retry_backoff(attempt_number)
        else:
            raise RuntimeError(
                f"Unsupported type for retry_backoff in task {self.task_identifier}"
            )

    def send(self, *args: P.args, **kwargs: P.kwargs) -> DurableTaskRun:
        """
        Send this task to the Hyrex queue with the provided parameters.

        This method accepts the same parameters as the wrapped function.
        """
        self.logger.debug(
            f"Sending task {self.func.__name__} to queue: {self.task_config.queue}"
        )

        # Convert positional args to kwargs based on function signature
        bound_args = self.signature.bind_partial(*args, **kwargs)
        bound_args.apply_defaults()
        
        # Validate the combined args against our function signature
        validated_kwargs = self._validate_kwargs(bound_args.arguments)

        current_context = get_hyrex_context()

        task_id = uuid7()
        task = EnqueueTaskRequest(
            id=task_id,
            durable_id=task_id,
            root_id=current_context.root_id if current_context else task_id,
            parent_id=current_context.task_id if current_context else None,
            task_name=self.task_identifier,
            queue=self.task_config.get_queue_name(),
            args=validated_kwargs,  # Pass the validated kwargs directly
            max_retries=self.task_config.max_retries,
            timeout_seconds=self.task_config.timeout_seconds,
            priority=self.task_config.priority,
            idempotency_key=self.task_config.idempotency_key,
            status=TaskStatus.queued,
            workflow_run_id=None,
            workflow_dependencies=None,
        )

        self.dispatcher.enqueue([task])

        return DurableTaskRun(
            task_name=self.task_identifier,
            durable_id=task.id,
            dispatcher=self.dispatcher,
        )

    def _validate_kwargs(self, kwargs):
        """
        Validate that the provided kwargs match the function signature.
        Apply type coercion if possible, otherwise raise appropriate errors.
        """
        validated_kwargs = {}

        # Check for missing required arguments
        for param_name, info in self.param_info.items():
            if param_name not in kwargs and not info.has_default:
                raise TypeError(
                    f"Missing required argument '{param_name}' for task '{self.task_identifier}'"
                )

        # Process provided arguments
        for param_name, value in kwargs.items():
            if param_name not in self.param_info:
                raise TypeError(
                    f"Unexpected argument '{param_name}' for task '{self.task_identifier}'"
                )

            param_type = self.param_info[param_name].type_hint

            # If we have a type hint, try to validate/convert the value
            if param_type is not None:
                try:
                    # Handle Pydantic models for backward compatibility
                    if hasattr(param_type, "model_validate"):
                        validated_kwargs[param_name] = param_type.model_validate(value)
                    elif hasattr(param_type, "parse_obj"):
                        validated_kwargs[param_name] = param_type.parse_obj(value)
                    # For primitive types, try basic conversion
                    elif param_type in (int, float, str, bool) and not isinstance(
                        value, param_type
                    ):
                        try:
                            validated_kwargs[param_name] = param_type(value)
                        except (ValueError, TypeError):
                            raise TypeError(
                                f"Cannot convert argument '{param_name}' value '{value}' to expected type {param_type.__name__}"
                            )
                    else:
                        # For other types, just pass the value as is
                        validated_kwargs[param_name] = value
                except Exception as e:
                    raise TypeError(
                        f"Validation error for argument '{param_name}': {str(e)}"
                    )
            else:
                # No type hint, just use the value as is
                validated_kwargs[param_name] = value

        # Add default values for missing arguments
        for param_name, info in self.param_info.items():
            if param_name not in kwargs and info.has_default:
                validated_kwargs[param_name] = info.default

        return validated_kwargs

    def get_arg_schema(self):
        """
        Return a schema describing the expected arguments for this task.
        This is useful for documentation and UI generation.
        """
        schema = {}
        for param_name, info in self.param_info.items():
            param_type = info.type_hint
            param_schema = {
                "required": not info.has_default,
                "type": param_type.__name__ if param_type else "any",
            }

            if info.has_default and info.default is not None:
                param_schema["default"] = str(info.default)

            # For Pydantic models, include their schema if available
            if param_type and hasattr(param_type, "model_json_schema"):
                param_schema["schema"] = param_type.model_json_schema()

            schema[param_name] = param_schema

        return schema

    def __repr__(self):
        return f"TaskWrapper<{self.task_identifier}>"

    def __call__(self, *args: P.args, **kwargs: P.kwargs):
        # Simply pass through all arguments to the original function
        return self.func(*args, **kwargs)

    # Methods for workflows:
    def __rshift__(self, other):
        builder = get_current_workflow_builder()
        if builder is None:
            raise RuntimeError(
                "No current workflow builder found. Please ensure you are within a workflow context."
            )
        # The builder is expected to have get_or_create_node.
        node = builder.get_or_create_node(self)
        return node >> other

    def __rrshift__(self, other):
        # This method is invoked when a TaskWrapper is on the right of >> and the left operand
        # does not implement __rshift__. We check if 'other' is a list.
        if isinstance(other, list):
            # Convert each element in the list to a DagNode (if needed)
            builder = get_current_workflow_builder()
            if builder is None:
                raise RuntimeError(
                    "No current workflow builder found. Please use a workflow context or decorator."
                )
            nodes = []
            for item in other:
                if isinstance(item, TaskWrapper):
                    # Note to developers: This path may be impossible?
                    node = builder.get_or_create_node(item)
                elif hasattr(item, "workflow_builder"):  # already a DagNode
                    node = item
                else:
                    raise TypeError(
                        f"Cannot use item of type {type(item)} as a task in the workflow."
                    )
                nodes.append(node)
            # Now, chain all the nodes in the list with self.
            # For example, we add self as a child to each node.
            for node in nodes:
                node >> self
            return self
        raise TypeError(f"Unsupported left operand type for >>: {type(other)}")



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: register_workflow.sql
import dataclasses
from typing import Any, Optional

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


REGISTER_WORKFLOW = """-- name: register_workflow \\:exec
INSERT INTO hyrex_workflow (workflow_name, cron_expr, source_code, dag_structure, last_updated)
VALUES (:p1, :p2, :p3, :p4, NOW())
ON CONFLICT (workflow_name)
DO UPDATE SET 
    cron_expr = EXCLUDED.cron_expr,
    source_code = EXCLUDED.source_code,
    dag_structure = EXCLUDED.dag_structure,
    last_updated = NOW()
"""


@dataclasses.dataclass()
class RegisterWorkflowParams:
    workflow_name: str
    cron_expr: Optional[str]
    source_code: Optional[str]
    dag_structure: Optional[Any]


class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def register_workflow(self, arg: RegisterWorkflowParams) -> None:
        self._conn.execute(sqlalchemy.text(REGISTER_WORKFLOW), {
            "p1": arg.workflow_name,
            "p2": arg.cron_expr,
            "p3": arg.source_code,
            "p4": arg.dag_structure,
        })


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def register_workflow(self, arg: RegisterWorkflowParams) -> None:
        await self._conn.execute(sqlalchemy.text(REGISTER_WORKFLOW), {
            "p1": arg.workflow_name,
            "p2": arg.cron_expr,
            "p3": arg.source_code,
            "p4": arg.dag_structure,
        })



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: get_executor_statuses.sql
import dataclasses
from typing import AsyncIterator, Iterator, List
import uuid

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


GET_EXECUTOR_STATUSES = """-- name: get_executor_statuses \\:many
SELECT
    id,
    status\\:\\:text as status
FROM hyrex_executor
WHERE id = ANY(:p1\\:\\:uuid[])
ORDER BY id
"""


@dataclasses.dataclass()
class GetExecutorStatusesParams:
    executor_ids: List[uuid.UUID]


@dataclasses.dataclass()
class GetExecutorStatusesRow:
    id: uuid.UUID
    status: str


class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def get_executor_statuses(self, arg: GetExecutorStatusesParams) -> Iterator[GetExecutorStatusesRow]:
        result = self._conn.execute(sqlalchemy.text(GET_EXECUTOR_STATUSES), {"p1": arg.executor_ids})
        for row in result:
            yield GetExecutorStatusesRow(
                id=row[0],
                status=row[1],
            )


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def get_executor_statuses(self, arg: GetExecutorStatusesParams) -> AsyncIterator[GetExecutorStatusesRow]:
        result = await self._conn.stream(sqlalchemy.text(GET_EXECUTOR_STATUSES), {"p1": arg.executor_ids})
        async for row in result:
            yield GetExecutorStatusesRow(
                id=row[0],
                status=row[1],
            )



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: fetch_result.sql
import dataclasses
from typing import Any, AsyncIterator, Iterator, Optional
import uuid

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


FETCH_RESULT = """-- name: fetch_result \\:many
SELECT result
FROM hyrex_task_result
WHERE task_id = :p1
"""


@dataclasses.dataclass()
class FetchResultParams:
    task_id: uuid.UUID


class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def fetch_result(self, arg: FetchResultParams) -> Iterator[Optional[Any]]:
        result = self._conn.execute(sqlalchemy.text(FETCH_RESULT), {"p1": arg.task_id})
        for row in result:
            yield row[0]


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def fetch_result(self, arg: FetchResultParams) -> AsyncIterator[Optional[Any]]:
        result = await self._conn.stream(sqlalchemy.text(FETCH_RESULT), {"p1": arg.task_id})
        async for row in result:
            yield row[0]



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: create_workflow_run.sql
import dataclasses
from typing import Any, Optional
import uuid

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


CREATE_WORKFLOW_RUN = """-- name: create_workflow_run \\:one
SELECT result FROM trigger_workflow_run(
    :p1\\:\\:UUID, 
    :p2, 
    :p3\\:\\:JSON, 
    :p4, 
    :p5, 
    :p6
) AS result
"""


@dataclasses.dataclass()
class CreateWorkflowRunParams:
    workflow_run_id: uuid.UUID
    workflow_name: str
    args: Any
    queue: str
    timeout_seconds: int
    idempotency_key: str


class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def create_workflow_run(self, arg: CreateWorkflowRunParams) -> Optional[Optional[Any]]:
        row = self._conn.execute(sqlalchemy.text(CREATE_WORKFLOW_RUN), {
            "p1": arg.workflow_run_id,
            "p2": arg.workflow_name,
            "p3": arg.args,
            "p4": arg.queue,
            "p5": arg.timeout_seconds,
            "p6": arg.idempotency_key,
        }).first()
        if row is None:
            return None
        return row[0]


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def create_workflow_run(self, arg: CreateWorkflowRunParams) -> Optional[Optional[Any]]:
        row = (await self._conn.execute(sqlalchemy.text(CREATE_WORKFLOW_RUN), {
            "p1": arg.workflow_run_id,
            "p2": arg.workflow_name,
            "p3": arg.args,
            "p4": arg.queue,
            "p5": arg.timeout_seconds,
            "p6": arg.idempotency_key,
        })).first()
        if row is None:
            return None
        return row[0]



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: create_cron_job_for_task.sql
import dataclasses
from typing import Optional

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


CREATE_CRON_JOB_FOR_TASK = """-- name: create_cron_job_for_task \\:exec
INSERT INTO hyrex_cron_job (schedule, command, jobname, job_source)
VALUES (:p1, :p2, :p3, 'TASK')
ON CONFLICT (jobname) 
DO UPDATE SET 
    schedule = EXCLUDED.schedule,
    command = EXCLUDED.command,
    job_source = 'TASK',
    active = true
"""


@dataclasses.dataclass()
class CreateCronJobForTaskParams:
    schedule: Optional[str]
    command: str
    jobname: str


class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def create_cron_job_for_task(self, arg: CreateCronJobForTaskParams) -> None:
        self._conn.execute(sqlalchemy.text(CREATE_CRON_JOB_FOR_TASK), {"p1": arg.schedule, "p2": arg.command, "p3": arg.jobname})


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def create_cron_job_for_task(self, arg: CreateCronJobForTaskParams) -> None:
        await self._conn.execute(sqlalchemy.text(CREATE_CRON_JOB_FOR_TASK), {"p1": arg.schedule, "p2": arg.command, "p3": arg.jobname})



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: get_executor_by_id.sql
import dataclasses
import datetime
from typing import Any, List, Optional
import uuid

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


GET_EXECUTOR_BY_ID = """-- name: get_executor_by_id \\:one
SELECT 
    e.id,
    e.name,
    e.worker_name,
    e.queue_pattern,
    e.queues,
    e.started,
    e.stopped,
    e.last_heartbeat,
    e.stats,
    CASE
        WHEN e.stopped IS NOT NULL THEN 'SHUTDOWN'
        WHEN (EXTRACT(EPOCH FROM (CURRENT_TIMESTAMP - e.last_heartbeat)) * 1000) > (20 * 1000) THEN 'LOST'
        WHEN (EXTRACT(EPOCH FROM (CURRENT_TIMESTAMP - e.last_heartbeat)) * 1000) <= (20 * 1000) THEN 'RUNNING'
        ELSE 'UNKNOWN'
    END AS status
FROM public.hyrex_executor e
WHERE e.id = :p1
"""


@dataclasses.dataclass()
class GetExecutorByIdParams:
    id: uuid.UUID


@dataclasses.dataclass()
class GetExecutorByIdRow:
    id: uuid.UUID
    name: str
    worker_name: str
    queue_pattern: str
    queues: List[str]
    started: Optional[datetime.datetime]
    stopped: Optional[datetime.datetime]
    last_heartbeat: Optional[datetime.datetime]
    stats: Optional[Any]
    status: str


class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def get_executor_by_id(self, arg: GetExecutorByIdParams) -> Optional[GetExecutorByIdRow]:
        row = self._conn.execute(sqlalchemy.text(GET_EXECUTOR_BY_ID), {"p1": arg.id}).first()
        if row is None:
            return None
        return GetExecutorByIdRow(
            id=row[0],
            name=row[1],
            worker_name=row[2],
            queue_pattern=row[3],
            queues=row[4],
            started=row[5],
            stopped=row[6],
            last_heartbeat=row[7],
            stats=row[8],
            status=row[9],
        )


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def get_executor_by_id(self, arg: GetExecutorByIdParams) -> Optional[GetExecutorByIdRow]:
        row = (await self._conn.execute(sqlalchemy.text(GET_EXECUTOR_BY_ID), {"p1": arg.id})).first()
        if row is None:
            return None
        return GetExecutorByIdRow(
            id=row[0],
            name=row[1],
            worker_name=row[2],
            queue_pattern=row[3],
            queues=row[4],
            started=row[5],
            stopped=row[6],
            last_heartbeat=row[7],
            stats=row[8],
            status=row[9],
        )



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: create_conditionally_retry_task_func.sql
import dataclasses

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


CREATE_CONDITIONALLY_RETRY_TASK_FUNC = """-- name: create_conditionally_retry_task_func \\:exec
CREATE OR REPLACE FUNCTION conditionally_retry_task(
    p_existing_task_id UUID,
    p_new_task_id UUID,
    p_timeout_seconds INT DEFAULT NULL
) RETURNS TABLE (
    success BOOLEAN,
    message TEXT,
    new_task_id UUID
) AS $$
DECLARE
    v_task_record RECORD;
    v_retry_created BOOLEAN \\:= FALSE;
BEGIN
    -- Get the existing task details
    SELECT * INTO v_task_record
    FROM hyrex_task_run
    WHERE id = p_existing_task_id
    FOR UPDATE;
    
    IF v_task_record IS NULL THEN
        RETURN QUERY SELECT FALSE, 'Task not found', NULL\\:\\:UUID;
        RETURN;
    END IF;
    
    -- Check if retry is allowed
    IF v_task_record.attempt_number >= v_task_record.max_retries THEN
        -- Log max retries reached
        INSERT INTO hyrex_system_logs (id, timestamp, event_name, event_body)
        VALUES (
            gen_random_uuid(),
            CURRENT_TIMESTAMP,
            'task_max_retries_reached',
            json_build_object(
                'task_id', p_existing_task_id,
                'task_name', v_task_record.task_name,
                'attempt_number', v_task_record.attempt_number,
                'max_retries', v_task_record.max_retries
            )
        );
        
        RETURN QUERY SELECT FALSE, 'Max retries reached', NULL\\:\\:UUID;
        RETURN;
    END IF;
    
    -- Create the retry task
    INSERT INTO hyrex_task_run (
        id,
        durable_id,
        root_id,
        parent_id,
        workflow_run_id,
        workflow_dependencies,
        task_name,
        args,
        queue,
        max_retries,
        priority,
        timeout_seconds,
        status,
        attempt_number,
        queued,
        idempotency_key,
        scheduled_start
    )
    VALUES (
        p_new_task_id,
        v_task_record.durable_id,
        v_task_record.root_id,
        v_task_record.parent_id,
        v_task_record.workflow_run_id,
        v_task_record.workflow_dependencies,
        v_task_record.task_name,
        v_task_record.args,
        v_task_record.queue,
        v_task_record.max_retries,
        v_task_record.priority,
        COALESCE(p_timeout_seconds, v_task_record.timeout_seconds),
        'QUEUED'\\:\\:task_run_status,
        v_task_record.attempt_number + 1,
        CURRENT_TIMESTAMP,
        NULL,  -- No idempotency key for retries
        NULL   -- No scheduled start for retries
    );
    
    -- Log the retry
    INSERT INTO hyrex_system_logs (id, timestamp, event_name, event_body)
    VALUES (
        gen_random_uuid(),
        CURRENT_TIMESTAMP,
        'task_retry_created',
        json_build_object(
            'original_task_id', p_existing_task_id,
            'new_task_id', p_new_task_id,
            'task_name', v_task_record.task_name,
            'attempt_number', v_task_record.attempt_number + 1,
            'max_retries', v_task_record.max_retries
        )
    );
    
    RETURN QUERY SELECT TRUE, 'Retry task created successfully', p_new_task_id;
END;
$$ LANGUAGE plpgsql
"""


@dataclasses.dataclass()
class CreateConditionallyRetryTaskFuncParams:

    pass

class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def create_conditionally_retry_task_func(self, arg: CreateConditionallyRetryTaskFuncParams) -> None:
        self._conn.execute(sqlalchemy.text(CREATE_CONDITIONALLY_RETRY_TASK_FUNC))


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def create_conditionally_retry_task_func(self, arg: CreateConditionallyRetryTaskFuncParams) -> None:
        await self._conn.execute(sqlalchemy.text(CREATE_CONDITIONALLY_RETRY_TASK_FUNC))



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: trim_task_stats.sql
import dataclasses
import datetime

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


TRIM_TASK_STATS = """-- name: trim_task_stats \\:exec
DELETE FROM hyrex_stats_task_status_counts WHERE timepoint < :p1
"""


@dataclasses.dataclass()
class TrimTaskStatsParams:
    timepoint: datetime.datetime


class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def trim_task_stats(self, arg: TrimTaskStatsParams) -> None:
        self._conn.execute(sqlalchemy.text(TRIM_TASK_STATS), {"p1": arg.timepoint})


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def trim_task_stats(self, arg: TrimTaskStatsParams) -> None:
        await self._conn.execute(sqlalchemy.text(TRIM_TASK_STATS), {"p1": arg.timepoint})



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: kv_get_value.sql
import dataclasses
from typing import Optional

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


GET_VALUE = """-- name: get_value \\:one
SELECT value FROM hyrex_kv WHERE key = :p1
"""


@dataclasses.dataclass()
class GetValueParams:
    key: str


class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def get_value(self, arg: GetValueParams) -> Optional[str]:
        row = self._conn.execute(sqlalchemy.text(GET_VALUE), {"p1": arg.key}).first()
        if row is None:
            return None
        return row[0]


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def get_value(self, arg: GetValueParams) -> Optional[str]:
        row = (await self._conn.execute(sqlalchemy.text(GET_VALUE), {"p1": arg.key})).first()
        if row is None:
            return None
        return row[0]



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: get_workflow_by_name.sql
import dataclasses
from typing import Optional

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


GET_WORKFLOW_BY_NAME = """-- name: get_workflow_by_name \\:one
SELECT 
    workflow_name,
    cron_expr,
    source_code,
    dag_structure,
    last_updated
FROM hyrex_workflow
WHERE workflow_name = :p1
"""


@dataclasses.dataclass()
class GetWorkflowByNameParams:
    workflow_name: str


class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def get_workflow_by_name(self, arg: GetWorkflowByNameParams) -> Optional[models.HyrexWorkflow]:
        row = self._conn.execute(sqlalchemy.text(GET_WORKFLOW_BY_NAME), {"p1": arg.workflow_name}).first()
        if row is None:
            return None
        return models.HyrexWorkflow(
            workflow_name=row[0],
            cron_expr=row[1],
            source_code=row[2],
            dag_structure=row[3],
            last_updated=row[4],
        )


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def get_workflow_by_name(self, arg: GetWorkflowByNameParams) -> Optional[models.HyrexWorkflow]:
        row = (await self._conn.execute(sqlalchemy.text(GET_WORKFLOW_BY_NAME), {"p1": arg.workflow_name})).first()
        if row is None:
            return None
        return models.HyrexWorkflow(
            workflow_name=row[0],
            cron_expr=row[1],
            source_code=row[2],
            dag_structure=row[3],
            last_updated=row[4],
        )



from pydantic import BaseModel

from hyrex.hyrex_registry import HyrexRegistry


class HyrexAppInfo(BaseModel):
    name: str


class HyrexApp:
    """
    Handles task registration and config for running Hyrex workers.
    """

    def __init__(self, app_name: str = "HelloHyrex"):
        self.registry: HyrexRegistry = HyrexRegistry()
        self.app_info = HyrexAppInfo(name=app_name)

    def add_registry(self, registry: HyrexRegistry):
        self.registry.add_registry(registry)



from hyrex.dispatcher.dispatcher_provider import get_dispatcher


class HyrexKV:
    """A simple key-value store for Hyrex applications.
    
    Provides string-to-string key-value storage with a maximum value size of 1MB.
    Can be used anywhere in your Hyrex tasks or application code.
    
    Example:
        from hyrex import HyrexKV
        
        # Set a value
        HyrexKV.set("user:123", "John Doe")
        
        # Get a value
        name = HyrexKV.get("user:123")  # Returns "John Doe"
        
        # Delete a value
        HyrexKV.delete("user:123")
    """
    
    MAX_VALUE_SIZE = 1_048_576  # 1MB in bytes
    
    @staticmethod
    def set(key: str, value: str) -> None:
        """Store a key-value pair.
        
        Args:
            key: The key to store the value under
            value: The string value to store (max 1MB)
            
        Raises:
            ValueError: If the value exceeds 1MB
            TypeError: If key or value is not a string
        """
        if not isinstance(key, str):
            raise TypeError(f"Key must be a string, got {type(key).__name__}")
        if not isinstance(value, str):
            raise TypeError(f"Value must be a string, got {type(value).__name__}")
            
        # Check value size (UTF-8 encoded)
        value_size = len(value.encode('utf-8'))
        if value_size > HyrexKV.MAX_VALUE_SIZE:
            raise ValueError(
                f"Value exceeds maximum size of 1MB. "
                f"Got {value_size:,} bytes"
            )
        
        dispatcher = get_dispatcher()
        dispatcher.kv_set(key, value)
    
    @staticmethod
    def get(key: str, default: str | None = None) -> str | None:
        """Retrieve a value by key.
        
        Args:
            key: The key to look up
            default: Default value to return if key doesn't exist
            
        Returns:
            The stored value, or default if key doesn't exist
            
        Raises:
            TypeError: If key is not a string
        """
        if not isinstance(key, str):
            raise TypeError(f"Key must be a string, got {type(key).__name__}")
            
        dispatcher = get_dispatcher()
        value = dispatcher.kv_get(key)
        return value if value is not None else default
    
    @staticmethod
    def delete(key: str) -> None:
        """Delete a key-value pair.
        
        Args:
            key: The key to delete
            
        Raises:
            TypeError: If key is not a string
        """
        if not isinstance(key, str):
            raise TypeError(f"Key must be a string, got {type(key).__name__}")
            
        dispatcher = get_dispatcher()
        dispatcher.kv_delete(key)


# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: batch_update_heartbeat_on_executors.sql
import dataclasses
from typing import List
import uuid

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


BATCH_UPDATE_HEARTBEAT_ON_EXECUTORS = """-- name: batch_update_heartbeat_on_executors \\:exec
UPDATE hyrex_executor
SET last_heartbeat = NOW()
WHERE id = ANY (:p1\\:\\:uuid[])
"""


@dataclasses.dataclass()
class BatchUpdateHeartbeatOnExecutorsParams:
    executor_ids: List[uuid.UUID]


class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def batch_update_heartbeat_on_executors(self, arg: BatchUpdateHeartbeatOnExecutorsParams) -> None:
        self._conn.execute(sqlalchemy.text(BATCH_UPDATE_HEARTBEAT_ON_EXECUTORS), {"p1": arg.executor_ids})


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def batch_update_heartbeat_on_executors(self, arg: BatchUpdateHeartbeatOnExecutorsParams) -> None:
        await self._conn.execute(sqlalchemy.text(BATCH_UPDATE_HEARTBEAT_ON_EXECUTORS), {"p1": arg.executor_ids})



import os

from hyrex.dispatcher.performance_dispatcher import PerformanceDispatcher
from hyrex.env_vars import EnvVars

from .dispatcher import Dispatcher
from .sqlc_dispatcher import SqlcDispatcher

# Single global dispatcher instance
_global_dispatcher: Dispatcher | None = None


def get_dispatcher() -> Dispatcher:
    """
    Get or create a singleton dispatcher instance.

    The first call to this function determines which dispatcher type will be used
    for the entire process. Subsequent calls return the same instance.

    The dispatcher type is determined by environment variables:
    - If API_KEY is set: Uses PerformanceDispatcher (cloud-based dispatcher)
    - If DATABASE_URL is set: Uses SqlcDispatcher (self-hosted PostgreSQL)

    Returns:
        A Dispatcher instance, reusing the global instance if one exists.

    Raises:
        ValueError: If neither API_KEY nor DATABASE_URL environment variables are set.
    """
    global _global_dispatcher

    # If we already have a global dispatcher, return it
    if _global_dispatcher is not None:
        return _global_dispatcher

    # Create the appropriate dispatcher based on environment variables
    api_key = os.environ.get(EnvVars.API_KEY)
    conn_string = os.environ.get(EnvVars.DATABASE_URL)

    if api_key:
        _global_dispatcher = PerformanceDispatcher(
            api_key=api_key, conn_string=conn_string
        )
    elif conn_string:
        _global_dispatcher = SqlcDispatcher(conn_string=conn_string)
    else:
        raise ValueError(
            f"Hyrex requires either {EnvVars.DATABASE_URL} or {EnvVars.API_KEY} to be set."
        )

    return _global_dispatcher



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: create_task_run_status_enum.sql
import dataclasses

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


CREATE_TASK_RUN_STATUS_ENUM = """-- name: create_task_run_status_enum \\:exec
DO $$
BEGIN
    IF NOT EXISTS (SELECT 1
                   FROM pg_type
                   WHERE typname = 'task_run_status'
                     AND typnamespace = 'public'\\:\\:regnamespace) THEN
        CREATE TYPE public.task_run_status AS ENUM (
            'SUCCESS',
            'FAILED',
            'RUNNING',
            'QUEUED',
            'UP_FOR_CANCEL',
            'CANCELED',
            'LOST',
            'STOPPED',
            'SKIPPED',
            'AWAIT_DEPS',
            'AWAIT_START_TIME'
            );
    END IF;
END $$
"""


@dataclasses.dataclass()
class CreateTaskRunStatusEnumParams:

    pass

class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def create_task_run_status_enum(self, arg: CreateTaskRunStatusEnumParams) -> None:
        self._conn.execute(sqlalchemy.text(CREATE_TASK_RUN_STATUS_ENUM))


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def create_task_run_status_enum(self, arg: CreateTaskRunStatusEnumParams) -> None:
        await self._conn.execute(sqlalchemy.text(CREATE_TASK_RUN_STATUS_ENUM))



import logging
from enum import StrEnum


class LogLevel(StrEnum):
    DEBUG = "DEBUG"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"


def init_logging(log_level: str):
    logger = logging.getLogger("hyrex")
    
    # Clear any existing handlers to prevent duplicates
    logger.handlers.clear()
    
    # Add our PID-formatted handler
    handler = logging.StreamHandler()
    handler.setFormatter(
        logging.Formatter(
            "[PID: %(process)d] %(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )
    )
    logger.addHandler(handler)
    logger.setLevel(level=getattr(logging, log_level.upper()))
    
    # Prevent propagation to root logger to avoid duplicate logs
    logger.propagate = False



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: create_hype_cron_job_run_details.sql
import dataclasses
import datetime
from typing import Any, Optional

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


CREATE_HYPE_CRON_JOB_RUN_DETAILS = """-- name: create_hype_cron_job_run_details \\:exec
INSERT INTO hype_cron_job_run_details (jobid, command_type, command_params, status, schedule_time)
VALUES (:p1, :p2, :p3, :p4, :p5)
"""


@dataclasses.dataclass()
class CreateHypeCronJobRunDetailsParams:
    jobid: int
    command_type: Any
    command_params: Any
    status: Optional[Any]
    schedule_time: datetime.datetime


class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def create_hype_cron_job_run_details(self, arg: CreateHypeCronJobRunDetailsParams) -> None:
        self._conn.execute(sqlalchemy.text(CREATE_HYPE_CRON_JOB_RUN_DETAILS), {
            "p1": arg.jobid,
            "p2": arg.command_type,
            "p3": arg.command_params,
            "p4": arg.status,
            "p5": arg.schedule_time,
        })


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def create_hype_cron_job_run_details(self, arg: CreateHypeCronJobRunDetailsParams) -> None:
        await self._conn.execute(sqlalchemy.text(CREATE_HYPE_CRON_JOB_RUN_DETAILS), {
            "p1": arg.jobid,
            "p2": arg.command_type,
            "p3": arg.command_params,
            "p4": arg.status,
            "p5": arg.schedule_time,
        })



from pydantic import BaseModel


class HyrexQueue(BaseModel):
    name: str
    concurrency_limit: int = 0

    def equals(self, other_queue: "HyrexQueue") -> bool:
        return (
            self.name == other_queue.name
            and self.concurrency_limit == other_queue.concurrency_limit
        )



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: _14_create_results_table.sql
import dataclasses

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


CREATE_RESULTS_TABLE = """-- name: create_results_table \\:exec
CREATE TABLE IF NOT EXISTS hyrex_task_result
(
    task_id    UUID PRIMARY KEY REFERENCES public.hyrex_task_run (id) ON DELETE CASCADE,
    result     JSON,
    created_at TIMESTAMP WITH TIME ZONE DEFAULT CURRENT_TIMESTAMP
)
"""


@dataclasses.dataclass()
class CreateResultsTableParams:

    pass

class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def create_results_table(self, arg: CreateResultsTableParams) -> None:
        self._conn.execute(sqlalchemy.text(CREATE_RESULTS_TABLE))


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def create_results_table(self, arg: CreateResultsTableParams) -> None:
        await self._conn.execute(sqlalchemy.text(CREATE_RESULTS_TABLE))



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: create_advance_workflow_run_func.sql
import dataclasses

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


CREATE_ADVANCE_WORKFLOW_RUN_FUNCTION = """-- name: create_advance_workflow_run_function \\:exec
CREATE OR REPLACE FUNCTION advance_workflow_run(p_workflow_run_id uuid)
    RETURNS TABLE (task_id uuid, new_status task_run_status, workflow_run_id uuid)
    LANGUAGE plpgsql
AS $$
BEGIN
    RETURN QUERY
        WITH latest AS (
            -- one row (the most-recent attempt) per durable_id in this workflow run
            SELECT DISTINCT ON (tr.durable_id)
                tr.id, tr.durable_id, tr.status, tr.workflow_dependencies
            FROM   hyrex_task_run tr
            WHERE  tr.workflow_run_id = p_workflow_run_id
            ORDER  BY tr.durable_id, tr.attempt_number DESC
            ),

            -- explode dependencies; keep rows even when there are none
            deps AS (
                SELECT l.id                        AS task_id,
                       d.dep_durable_id
                FROM   latest l
                           LEFT   JOIN LATERAL unnest(l.workflow_dependencies)  -- returns zero rows for NULL/empty
                    AS d(dep_durable_id) ON TRUE
                ),

            dep_status AS (
                /* For every dependency edge, pull the latest status of that dependency.
                   If a dependency durable_id doesn’t exist in "latest", dep_status will be NULL. */
                SELECT d.task_id,
                       bool_and(ld.status = 'SUCCESS')          AS all_success,
                       bool_and(ld.status IS NOT NULL)          AS all_present
                FROM   deps d
                           LEFT   JOIN latest ld ON ld.durable_id = d.dep_durable_id
                GROUP BY d.task_id
                ),

            ready AS (
                /* A task is ready when\\:
                   • its own latest status is AWAIT_DEPS, AND
                   • (it has no deps) OR (all deps are present AND all succeeded) */
                SELECT l.id
                FROM   latest l
                           LEFT   JOIN dep_status s ON s.task_id = l.id
                WHERE  l.status = 'AWAIT_DEPS'
                  AND (
                    l.workflow_dependencies IS NULL
                        OR array_length(l.workflow_dependencies,1) = 0
                        OR (s.all_present AND s.all_success)
                    )
                ),

            locked AS (
                -- make promotion idempotent under concurrency
                SELECT tr.id
                FROM   hyrex_task_run tr
                           JOIN   ready r ON r.id = tr.id
                    FOR UPDATE SKIP LOCKED
                )

            UPDATE hyrex_task_run trg
                SET status = 'QUEUED'\\:\\:task_run_status,
                    queued = now()
                FROM locked
                WHERE trg.id = locked.id
                RETURNING trg.id AS task_id, trg.status AS status, trg.workflow_run_id AS workflow_run_id;
END;
$$
"""


@dataclasses.dataclass()
class CreateAdvanceWorkflowRunFunctionParams:

    pass

class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def create_advance_workflow_run_function(self, arg: CreateAdvanceWorkflowRunFunctionParams) -> None:
        self._conn.execute(sqlalchemy.text(CREATE_ADVANCE_WORKFLOW_RUN_FUNCTION))


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def create_advance_workflow_run_function(self, arg: CreateAdvanceWorkflowRunFunctionParams) -> None:
        await self._conn.execute(sqlalchemy.text(CREATE_ADVANCE_WORKFLOW_RUN_FUNCTION))



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: create_workflow_trigger_func.sql
import dataclasses

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


CREATE_WORKFLOW_TRIGGER = """-- name: create_workflow_trigger \\:exec

CREATE OR REPLACE FUNCTION trigger_workflow_run(
    p_workflow_run_id UUID,
    p_workflow_name TEXT,
    p_args JSON,
    p_queue TEXT,
    p_timeout_seconds INTEGER,
    p_idempotency_key TEXT
) RETURNS TABLE (
    workflow_run_id UUID,
    task_count INTEGER,
    success BOOLEAN,
    message TEXT
) AS $$
DECLARE
    v_workflow_record RECORD;
    v_node RECORD;
    v_task_id UUID;
    v_task_count INTEGER \\:= 0;
    v_workflow_dependencies UUID[];
    v_dep_durable_id TEXT;
    v_node_id_map JSONB \\:= '{}'\\:\\:JSONB;
BEGIN
    -- Validate timeout_seconds
    IF p_timeout_seconds IS NOT NULL AND p_timeout_seconds <= 0 THEN
        RETURN QUERY SELECT
            NULL\\:\\:UUID,
            0,
            FALSE,
            'Invalid timeout_seconds\\: must be NULL or greater than 0';
        RETURN;
    END IF;

    -- Start transaction
    BEGIN
        -- Get workflow definition including DAG structure
        SELECT * INTO v_workflow_record
        FROM hyrex_workflow
        WHERE workflow_name = p_workflow_name
        LIMIT 1;

        IF v_workflow_record IS NULL THEN
            RETURN QUERY SELECT
                NULL\\:\\:UUID,
                0,
                FALSE,
                'Workflow not found\\: ' || p_workflow_name;
            RETURN;
        END IF;

        -- Insert workflow run
        INSERT INTO hyrex_workflow_run (
            id,
            parent_id,
            workflow_name,
            args,
            queue,
            timeout_seconds,
            status,
            queued,
            last_heartbeat,
            idempotency_key
        )
        VALUES (
            p_workflow_run_id,
            NULL,
            p_workflow_name,
            p_args,
            p_queue,
            p_timeout_seconds,
            'RUNNING'\\:\\:workflow_run_status,
            NOW(),
            NOW(),
            p_idempotency_key
        );

        -- Create task runs from DAG structure
        IF v_workflow_record.dag_structure IS NOT NULL AND
           v_workflow_record.dag_structure->'nodes' IS NOT NULL THEN

            -- First pass\\: Create mapping of node IDs to task UUIDs
            FOR v_node IN
                SELECT value FROM jsonb_array_elements(v_workflow_record.dag_structure->'nodes')
            LOOP
                -- Generate a new UUID for this task
                v_task_id \\:= gen_random_uuid();

                -- Store the mapping from node ID to task UUID
                v_node_id_map \\:= v_node_id_map || jsonb_build_object(v_node.value->>'id', v_task_id\\:\\:TEXT);
            END LOOP;

            -- Second pass\\: Create task runs with proper dependencies
            FOR v_node IN
                SELECT value FROM jsonb_array_elements(v_workflow_record.dag_structure->'nodes')
            LOOP
                -- Get the task UUID we generated for this node
                v_task_id \\:= (v_node_id_map->>(v_node.value->>'id'))\\:\\:UUID;

                -- Build workflow_dependencies array based on edges
                v_workflow_dependencies \\:= ARRAY[]\\:\\:UUID[];

                -- Find all nodes that point to this node (dependencies)
                FOR v_dep_durable_id IN
                    SELECT DISTINCT value->>'from'
                    FROM jsonb_array_elements(v_workflow_record.dag_structure->'edges')
                    WHERE value->>'to' = v_node.value->>'id'
                LOOP
                    -- Map the dependency node ID to its task UUID
                    v_workflow_dependencies \\:= array_append(v_workflow_dependencies, (v_node_id_map->>v_dep_durable_id)\\:\\:UUID);
                END LOOP;

                -- Use create_task_run function with workflow parameters
                PERFORM create_task_run(
                    v_task_id,
                    v_task_id,  -- SDK uses same UUID for id and durable_id
                    v_task_id,  -- SDK uses same UUID for root_id too
                    NULL\\:\\:UUID,  -- No parent for workflow tasks
                    CASE
                        WHEN array_length(v_workflow_dependencies, 1) IS NULL OR array_length(v_workflow_dependencies, 1) = 0
                        THEN 'QUEUED'\\:\\:task_run_status
                        ELSE 'AWAIT_DEPS'\\:\\:task_run_status
                    END,
                    v_node.value->>'name',
                    p_args,  -- Pass workflow args to each task
                    p_queue,
                    3\\:\\:SMALLINT,  -- Default max retries
                    0\\:\\:SMALLINT,  -- Default priority
                    p_timeout_seconds,
                    NULL\\:\\:VARCHAR,  -- No idempotency key for individual tasks
                    NULL\\:\\:TIMESTAMP WITH TIME ZONE,  -- No scheduled_start for workflow tasks
                    p_workflow_run_id,
                    v_workflow_dependencies
                );

                v_task_count \\:= v_task_count + 1;
            END LOOP;
        END IF;

        -- Return success
        RETURN QUERY SELECT
            p_workflow_run_id,
            v_task_count,
            TRUE,
            'Workflow triggered successfully with ' || v_task_count || ' tasks';

    EXCEPTION WHEN OTHERS THEN
        -- Rollback will happen automatically
        RETURN QUERY SELECT
            NULL\\:\\:UUID,
            0,
            FALSE,
            'Error triggering workflow\\: ' || SQLERRM;
    END;
END;
$$ LANGUAGE plpgsql
"""


@dataclasses.dataclass()
class CreateWorkflowTriggerParams:

    pass

class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def create_workflow_trigger(self, arg: CreateWorkflowTriggerParams) -> None:
        self._conn.execute(sqlalchemy.text(CREATE_WORKFLOW_TRIGGER))


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def create_workflow_trigger(self, arg: CreateWorkflowTriggerParams) -> None:
        await self._conn.execute(sqlalchemy.text(CREATE_WORKFLOW_TRIGGER))



import logging
import time
from typing import Any, Optional
from uuid import UUID

from pydantic import BaseModel, PrivateAttr

from hyrex.dispatcher.dispatcher import Dispatcher
from hyrex.dispatcher.dispatcher_provider import get_dispatcher
from hyrex.schemas import TaskRun, TaskStatus


class DurableTaskRun(BaseModel):
    task_name: str
    durable_id: UUID
    task_runs: list[TaskRun] = []

    _dispatcher: Dispatcher = PrivateAttr(default_factory=get_dispatcher)
    _logger: logging.Logger = PrivateAttr(
        default_factory=lambda: logging.getLogger(__name__)
    )

    def wait(self, timeout: float = 30.0, interval: float = 0.5) -> bool:
        start = time.time()
        elapsed = 0

        run_complete = False

        while not run_complete:
            self.refresh()
            for task in self.task_runs:
                if task.status == TaskStatus.success:
                    # Completed successfully
                    return True
                elif (
                    task.status == TaskStatus.failed
                    and task.attempt_number == task.max_retries
                ):
                    # Failed with no retries left
                    return False
                elif task.status == TaskStatus.canceled:
                    # Canceled
                    return False

            time.sleep(interval)
            elapsed = time.time() - start
            if elapsed > timeout:
                raise TimeoutError("Waiting for durable task run timed out.")

    def get_result(self):
        self.refresh()
        for task in self.task_runs:
            if task.status == TaskStatus.success and task.result is not None:
                return task.result
        self._logger.warning(f"No result found for durable run {self.durable_id}.")
        return None

    def cancel(self):
        self._dispatcher.try_to_cancel_durable_run(self.durable_id)

    def __repr__(self):
        return f"DurableTaskRun<{self.task_name}>[{self.durable_id}]"

    def refresh(self):
        self.task_runs = self._dispatcher.get_durable_run_tasks(self.durable_id)

        # Update task_name from the first task run
        if self.task_runs and len(self.task_runs) > 0:
            self.task_name = self.task_runs[0].task_name



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: _16_create_hype_cron_job_table.sql
import dataclasses

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


CREATE_HYPE_CRON_JOB_TABLE = """-- name: create_hype_cron_job_table \\:exec
CREATE TABLE IF NOT EXISTS hype_cron_job
(
    jobid                          bigserial PRIMARY KEY,
    schedule                       text,
    command_type                   hype_command_type NOT NULL,
    command_params                 JSONB   NOT NULL,
    active                         boolean NOT NULL DEFAULT true,
    jobname                        text    NOT NULL,
    activated_at                   timestamptz      default now(),
    scheduled_jobs_confirmed_until timestamptz      default now(),
    should_backfill                boolean default true,
    UNIQUE (jobname)
)
"""


@dataclasses.dataclass()
class CreateHypeCronJobTableParams:

    pass

class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def create_hype_cron_job_table(self, arg: CreateHypeCronJobTableParams) -> None:
        self._conn.execute(sqlalchemy.text(CREATE_HYPE_CRON_JOB_TABLE))


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def create_hype_cron_job_table(self, arg: CreateHypeCronJobTableParams) -> None:
        await self._conn.execute(sqlalchemy.text(CREATE_HYPE_CRON_JOB_TABLE))
